# Copyright (c) 2022, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil

import pytest
import torch
from lightning.pytorch import Trainer
from omegaconf import OmegaConf

from nemo.core import ModelPT
from nemo.utils import logging
from nemo.utils.exp_manager import CallbackParams, ExpManagerConfig, StatelessTimer, exp_manager


class OnesDataset(torch.utils.data.Dataset):
    def __init__(self, dataset_len):
        super().__init__()
        self.__dataset_len = dataset_len

    def __getitem__(self, *args):
        return torch.ones(2)

    def __len__(self):
        return self.__dataset_len


class ExampleModel(ModelPT):
    def __init__(self, *args, **kwargs):
        cfg = OmegaConf.structured({})
        super().__init__(cfg, trainer=kwargs.get('trainer', None))
        # dummy parameter in order to allow DDP to execute
        self.l1 = torch.nn.modules.Linear(in_features=2, out_features=1)

    def train_dataloader(self):
        dataset = OnesDataset(10000)
        return torch.utils.data.DataLoader(dataset, batch_size=2, num_workers=4)

    def val_dataloader(self):
        dataset = OnesDataset(10)
        return torch.utils.data.DataLoader(dataset, batch_size=2, num_workers=4)

    def predict_dataloader(self):
        dataset = OnesDataset(10)
        return torch.utils.data.DataLoader(dataset, batch_size=2, num_workers=4)

    def forward(self, batch):
        return (self.l1(batch) - batch.mean(dim=1)).mean()

    def validation_step(self, batch, batch_idx):
        loss = (self.l1(batch) - batch.mean(dim=1)).mean()
        self.validation_step_outputs.append(loss)
        return loss

    def training_step(self, batch, batch_idx):
        return (self.l1(batch) - batch.mean(dim=1)).mean()

    def list_available_models(self):
        pass

    def setup_training_data(self):
        pass

    def setup_validation_data(self):
        pass

    def on_validation_epoch_end(self):
        if not self.validation_step_outputs:
            return
        self.log("val_loss", torch.stack(self.validation_step_outputs).mean(), sync_dist=True)
        self.validation_step_outputs.clear()  # free memory


class TestStatelessTimer:
    def setup_model(self):
        # Stateless timer for 3 seconds.
        # Max steps shouldn't matter for it should stop in 3 seconds based on the timer.
        # Val check interval makes sure a checkpoint is written and can be restored from.
        callback_params = CallbackParams()
        callback_params.monitor = "val_loss"
        callback_params.save_top_k = 1
        trainer = Trainer(
            devices=1,
            val_check_interval=5,
            max_steps=10000,
            accelerator='gpu',
            strategy='ddp',
            logger=False,
            enable_checkpointing=False,
        )
        exp_manager_cfg = ExpManagerConfig(
            explicit_log_dir='./ptl_stateless_timer_check/',
            use_datetime_version=False,
            version="",
            resume_ignore_no_checkpoint=True,
            create_checkpoint_callback=True,
            checkpoint_callback_params=callback_params,
            resume_if_exists=True,
            max_time_per_run="00:00:00:03",
        )
        exp_manager(trainer, cfg=OmegaConf.structured(exp_manager_cfg))
        model = ExampleModel(trainer=trainer)
        trainer.fit(model)
        return trainer

    def cleanup(self):
        if os.path.exists('./ptl_stateless_timer_check'):
            shutil.rmtree('./ptl_stateless_timer_check', ignore_errors=True)

    @pytest.mark.run_only_on('GPU')
    @pytest.mark.unit
    def test_stateless_timer(self):
        self.cleanup()
        trainer = self.setup_model()
        global_step_1 = trainer.global_step
        trainer = self.setup_model()
        global_step_2 = trainer.global_step
        trainer = self.setup_model()
        global_step_3 = trainer.global_step
        logging.info(f"Global steps : {global_step_1}, {global_step_2}, {global_step_3}")
        assert global_step_3 > global_step_2 > global_step_1
        self.cleanup()



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import defaultdict

import torch
from whisper_normalizer.english import EnglishTextNormalizer

from nemo.collections.asr.metrics.wer import word_error_rate
from nemo.utils import logging


class WER:
    """
    Computes WER on text predictions.
    By default, uses Whisper's EnglishTextNormalizer on hypotheses and references.
    """

    def __init__(self, normalize: bool = True, normalizer=None, verbose: bool = True):
        self.verbose = verbose
        if normalize:
            if normalizer is None:
                self.normalizer = EnglishTextNormalizer()
            else:
                self.normalizer = normalizer
        else:
            self.normalizer = _identity

        self._refs = defaultdict(list)
        self._hyps = defaultdict(list)

    def reset(self):
        self._refs.clear()
        self._hyps.clear()
        return self

    def update(self, name: str, refs: list[str], hyps: list[str]) -> None:
        for ref, hyp in zip(refs, hyps):
            self._refs[name].append(self.normalizer(ref))
            self._hyps[name].append(self.normalizer(hyp))
        if self.verbose and refs and hyps:
            logging.info(f"[REF]\t{refs[0]}\n[HYP]\t{hyps[0]}")

    def compute(self) -> dict[str, torch.Tensor]:
        corpus_metric = {}
        for name in self._refs.keys():
            metric = torch.tensor(word_error_rate(self._hyps[name], self._refs[name]))
            corpus_metric[f"wer_{name}"] = metric
        corpus_metric["wer"] = torch.stack(list(corpus_metric.values())).mean()
        self.reset()
        return corpus_metric


def _identity(x):
    return x



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import nemo_run as run
import pytest

from nemo.collections.llm.api import finetune, pretrain
from nemo.collections.vlm import Gemma3VLConfig12B, Gemma3VLModel, LoRA
from nemo.collections.vlm.recipes import gemma3vl_12b
from nemo.lightning import Trainer


class TestGemma3VL12B:
    @pytest.fixture(scope="class")
    def recipe_module(self):
        return gemma3vl_12b

    def test_model(self, recipe_module):
        model_config = recipe_module.model()
        # Check that the model configuration is a run.Config instance wrapping the Gemma3VLModel
        assert isinstance(model_config, run.Config)
        # Verify that the factory function is the Gemma3VLModel
        assert model_config.__fn_or_cls__ == Gemma3VLModel
        # Verify the inner configuration is a run.Config for Gemma3VLConfig12B
        assert isinstance(model_config.config, run.Config)
        assert model_config.config.__fn_or_cls__ == Gemma3VLConfig12B

    def test_pretrain_recipe_default(self, recipe_module):
        recipe = recipe_module.pretrain_recipe()
        # Check that the returned recipe is a run.Partial wrapping pretrain
        assert isinstance(recipe, run.Partial)
        assert recipe.__fn_or_cls__ == pretrain

        # Verify the model is correctly set
        assert isinstance(recipe.model, run.Config)
        assert recipe.model.__fn_or_cls__ == Gemma3VLModel

        # Verify trainer configuration
        assert isinstance(recipe.trainer, run.Config)
        # Trainer should be the one from nemo.lightning (Trainer)
        assert recipe.trainer.__fn_or_cls__ == Trainer
        assert recipe.trainer.accelerator == "gpu"
        # Default values: num_nodes=8 and num_gpus_per_node=8
        assert recipe.trainer.num_nodes == 8
        assert recipe.trainer.devices == 8

        # Verify strategy settings
        strat = recipe.trainer.strategy
        assert isinstance(strat, run.Config)
        assert strat.tensor_model_parallel_size == 4
        assert strat.pipeline_model_parallel_size == 1
        assert strat.sequence_parallel is True

        # Verify data configuration
        assert isinstance(recipe.data, run.Config)
        # The recipe uses the Gemma3VLMockDataModule with the following parameters:
        assert recipe.data.__fn_or_cls__.__name__ == "Gemma3VLMockDataModule"
        assert recipe.data.seq_length == 512
        assert recipe.data.global_batch_size == 512
        assert recipe.data.micro_batch_size == 1

        # Verify logging and resume configurations are set (non-null)
        assert recipe.log is not None
        assert recipe.resume is not None

    def test_finetune_recipe_default(self, recipe_module):
        recipe = recipe_module.finetune_recipe()
        # Check that the returned recipe is a run.Partial wrapping finetune
        assert isinstance(recipe, run.Partial)
        assert recipe.__fn_or_cls__ == finetune

        # Verify the model is correctly set
        assert isinstance(recipe.model, run.Config)
        assert recipe.model.__fn_or_cls__ == Gemma3VLModel

        # Verify trainer configuration
        assert isinstance(recipe.trainer, run.Config)
        # Trainer should be the one from nemo.lightning (Trainer)
        assert recipe.trainer.__fn_or_cls__ == Trainer
        assert recipe.trainer.accelerator == "gpu"
        # Default values: num_nodes=8 and num_gpus_per_node=8
        assert recipe.trainer.num_nodes == 8
        assert recipe.trainer.devices == 8

        # Verify strategy settings (note: for 'none' peft, tensor_model_parallel_size is updated to 4)
        strat = recipe.trainer.strategy
        assert isinstance(strat, run.Config)
        assert strat.tensor_model_parallel_size == 4
        assert strat.pipeline_model_parallel_size == 1
        # encoder_pipeline_model_parallel_size is set to 0 and sequence_parallel should be True
        assert strat.encoder_pipeline_model_parallel_size == 0
        assert strat.sequence_parallel is True

        # Verify data configuration
        assert isinstance(recipe.data, run.Config)
        # The recipe uses the Gemma3VLMockDataModule with the following parameters:
        assert recipe.data.__fn_or_cls__.__name__ == "Gemma3VLMockDataModule"
        assert recipe.data.seq_length == 512
        assert recipe.data.global_batch_size == 32
        assert recipe.data.micro_batch_size == 1
        assert recipe.data.num_workers == 4

        # Verify logging and resume configurations are set (non-null)
        assert recipe.log is not None
        assert recipe.resume is not None

    def test_finetune_recipe_peft_lora(self, recipe_module):
        # Test the fine-tuning recipe with peft_scheme set to "lora"
        recipe = recipe_module.finetune_recipe(peft_scheme="lora")
        # In this case, a peft field should be present and configured for LoRA
        assert hasattr(recipe, "peft")
        assert isinstance(recipe.peft, run.Config)
        assert recipe.peft.__fn_or_cls__ == LoRA

        # The learning rate should have been updated for LoRA usage
        assert recipe.optim.config.lr == 1e-4

    @pytest.mark.parametrize("num_nodes,num_gpus", [(1, 8), (2, 4)])
    def test_finetune_recipe_different_configurations(self, recipe_module, num_nodes, num_gpus):
        # Verify that the recipe honors different numbers of nodes and GPUs per node
        recipe = recipe_module.finetune_recipe(num_nodes=num_nodes, num_gpus_per_node=num_gpus)
        assert recipe.trainer.num_nodes == num_nodes
        assert recipe.trainer.devices == num_gpus

    def test_pretrain_performance_optimizations(self, recipe_module):
        # Test the performance optimizations for pre-training
        recipe = recipe_module.pretrain_recipe(performance_mode=True)

        # Verify that performance-related callbacks are added
        callbacks = recipe.trainer.callbacks
        assert len(callbacks) > 1  # Should have at least TimingCallback plus performance callbacks

        # Verify that grad_reduce_in_fp32 is set to False
        assert recipe.trainer.plugins.grad_reduce_in_fp32 is False

    def test_pretrain_recipe_different_configurations(self, recipe_module):
        # Test different configurations for pre-training
        recipe = recipe_module.pretrain_recipe(num_nodes=2, num_gpus_per_node=4)
        assert recipe.trainer.num_nodes == 2
        assert recipe.trainer.devices == 4



# Copyright (c) 2022, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import torch

from nemo.collections.tts.parts.utils.helpers import regulate_len, sort_tensor, unsort_tensor


def sample_duration_input(max_length=64, group_size=2, batch_size=3):
    generator = torch.Generator()
    generator.manual_seed(0)
    lengths = torch.randint(max_length // 4, max_length - 7, (batch_size,), generator=generator)
    durs = torch.ones(batch_size, max_length) * group_size
    durs[0, lengths[0]] += 1
    durs[2, lengths[2]] -= 1
    enc = torch.randint(16, 64, (batch_size, max_length, 17))
    return durs, enc, lengths


@pytest.mark.unit
def test_sort_unsort():
    durs_in, enc_in, dur_lens = sample_duration_input(batch_size=13)
    print("In: ", enc_in)
    sorted_enc, sorted_len, sorted_ids = sort_tensor(enc_in, dur_lens)
    unsorted_enc = unsort_tensor(sorted_enc, sorted_ids)
    print("Out: ", unsorted_enc)

    assert torch.all(unsorted_enc == enc_in)


@pytest.mark.unit
def test_regulate_len():
    group_size = 2
    durs_in, enc_in, dur_lens = sample_duration_input(group_size=group_size)
    enc_out, lens_out = regulate_len(durs_in, enc_in, group_size=group_size, dur_lens=dur_lens)
    # make sure lens_out are rounded
    sum_diff = lens_out - torch.mul(lens_out // group_size, group_size)
    assert sum_diff.sum(dim=0) == 0
    # make sure all round-ups are <= group_size
    diff = lens_out - durs_in.sum(dim=1)
    assert torch.max(diff) < group_size



# Copyright (c) 2023, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import torch

from nemo.collections.asr.models.configs.asr_models_config import CacheAwareStreamingConfig
from nemo.collections.asr.parts.utils.multispk_transcribe_utils import MultiTalkerInstanceManager
from tests.collections.asr.test_asr_rnnt_encoder_model_bpe import asr_model as offline_asr_model
from tests.collections.speaker_tasks.test_diar_sortformer_models import sortformer_model as diar_model


@pytest.fixture()
def asr_model(offline_asr_model):
    """Wrapper fixture that adds streaming_cfg to the asr_model from test_asr_rnnt_encoder_model_bpe"""
    # Add streaming_cfg to encoder for streaming tests
    streaming_cfg = CacheAwareStreamingConfig(
        valid_out_len=1,
        drop_extra_pre_encoded=7,
        chunk_size=8,
        shift_size=4,
        cache_drop_size=4,
        last_channel_cache_size=64,
        pre_encode_cache_size=0,
        last_channel_num=0,
        last_time_num=0,
    )
    offline_asr_model.encoder.streaming_cfg = streaming_cfg

    # Mock get_initial_cache_state method for MultiTalkerInstanceManager tests
    def get_initial_cache_state(batch_size=1):
        """Mock method to return initial cache state for streaming"""
        # Return dummy cache state tensors
        cache_last_channel = torch.zeros(2, batch_size, 64)
        cache_last_time = torch.zeros(2, batch_size, 64)
        cache_last_channel_len = torch.zeros(batch_size)
        return (cache_last_channel, cache_last_time, cache_last_channel_len)

    offline_asr_model.encoder.get_initial_cache_state = get_initial_cache_state

    return offline_asr_model


class TestMultiTalkerInstanceManagerMethods:
    """Test methods of the MultiTalkerInstanceManager class"""

    @pytest.mark.unit
    def test_reset_active_speaker_buffers(self, asr_model, diar_model):
        """Test _reset_active_speaker_buffers method"""
        instance_manager = MultiTalkerInstanceManager(
            asr_model=asr_model,
            diar_model=diar_model,
            batch_size=2,
            max_num_of_spks=4,
            sent_break_sec=5.0,
        )

        # Populate some buffers first
        # pylint: disable=protected-access
        instance_manager._active_chunk_audio = [torch.randn(100)]
        instance_manager._active_chunk_lengths = [torch.tensor(100)]
        instance_manager._active_speaker_targets = [torch.randn(10)]

        # Reset the buffers
        instance_manager._reset_active_speaker_buffers()

        # Verify all buffers are empty
        assert len(instance_manager._active_chunk_audio) == 0
        assert len(instance_manager._active_chunk_lengths) == 0
        assert len(instance_manager._active_speaker_targets) == 0
        assert len(instance_manager._inactive_speaker_targets) == 0
        assert len(instance_manager._active_previous_hypotheses) == 0
        assert len(instance_manager._active_asr_pred_out_stream) == 0
        assert len(instance_manager._active_cache_last_channel) == 0
        assert len(instance_manager._active_cache_last_time) == 0
        assert len(instance_manager._active_cache_last_channel_len) == 0
        # pylint: enable=protected-access

    @pytest.mark.unit
    def test_reset_with_new_params(self, asr_model, diar_model):
        """Test reset method with new batch_size and max_num_of_spks"""
        instance_manager = MultiTalkerInstanceManager(
            asr_model=asr_model,
            diar_model=diar_model,
            batch_size=2,
            max_num_of_spks=4,
            sent_break_sec=5.0,
        )

        # Reset with new parameters
        instance_manager.reset(batch_size=3, max_num_of_spks=6)

        # Verify new parameters are applied
        assert instance_manager.batch_size == 3
        assert instance_manager.max_num_of_spks == 6
        assert len(instance_manager.batch_asr_states) == 3

    @pytest.mark.unit
    def test_add_speaker(self, asr_model, diar_model):
        """Test add_speaker method"""
        instance_manager = MultiTalkerInstanceManager(
            asr_model=asr_model,
            diar_model=diar_model,
            batch_size=2,
            max_num_of_spks=4,
            sent_break_sec=5.0,
        )
        instance_manager.reset()

        # Initially, batch 0 should have speaker [0]
        speakers_before = instance_manager.get_speakers(batch_idx=0)
        assert 0 in speakers_before

        # Add speaker 1
        instance_manager.add_speaker(batch_idx=0, speaker_id=1)

        # Verify speaker 1 is added
        speakers_after = instance_manager.get_speakers(batch_idx=0)
        assert 0 in speakers_after
        assert 1 in speakers_after

    @pytest.mark.unit
    def test_update_diar_state(self, asr_model, diar_model):
        """Test update_diar_state method"""
        instance_manager = MultiTalkerInstanceManager(
            asr_model=asr_model,
            diar_model=diar_model,
            batch_size=2,
            max_num_of_spks=4,
            sent_break_sec=5.0,
        )
        instance_manager.reset()

        # Create mock diarization data
        diar_pred_out_stream = torch.randn(2, 20, 4)
        previous_chunk_preds = torch.randn(2, 10, 4)

        # Get initial streaming state from diar_model
        diar_streaming_state = diar_model.sortformer_modules.init_streaming_state(batch_size=2)

        # Update diar state
        instance_manager.update_diar_state(
            diar_pred_out_stream=diar_pred_out_stream,
            previous_chunk_preds=previous_chunk_preds,
            diar_streaming_state=diar_streaming_state,
        )

        # Verify diar state is updated
        assert torch.equal(instance_manager.diar_states.diar_pred_out_stream, diar_pred_out_stream)
        assert torch.equal(instance_manager.diar_states.previous_chunk_preds, previous_chunk_preds)
        assert instance_manager.diar_states.streaming_state is not None

    @pytest.mark.unit
    def test_update_asr_state(self, asr_model, diar_model):
        """Test update_asr_state method"""
        instance_manager = MultiTalkerInstanceManager(
            asr_model=asr_model,
            diar_model=diar_model,
            batch_size=2,
            max_num_of_spks=4,
            sent_break_sec=5.0,
        )
        instance_manager.reset()

        # Get the initial cache state structure
        asr_state = instance_manager.batch_asr_states[0]

        # Create mock ASR cache data with correct shapes
        cache_shape = asr_state.cache_last_channel.shape
        time_shape = asr_state.cache_last_time.shape

        cache_last_channel = torch.randn(cache_shape[0], cache_shape[2])  # Remove speaker dimension
        cache_last_time = torch.randn(time_shape[0], time_shape[2])
        cache_last_channel_len = torch.tensor([10])

        # Create a simple mock hypothesis
        from nemo.collections.asr.parts.utils.rnnt_utils import Hypothesis

        previous_hypothesis = Hypothesis(score=0.0, y_sequence=[], text="test")
        previous_pred_out = torch.randn(1, 10, 128)

        # Update ASR state for batch 0, speaker 0
        instance_manager.update_asr_state(
            batch_idx=0,
            speaker_id=0,
            cache_last_channel=cache_last_channel,
            cache_last_time=cache_last_time,
            cache_last_channel_len=cache_last_channel_len,
            previous_hypotheses=previous_hypothesis,
            previous_pred_out=previous_pred_out,
        )

        # Verify the state was updated
        updated_asr_state = instance_manager.batch_asr_states[0]
        assert updated_asr_state.previous_hypothesis[0] is previous_hypothesis
        assert updated_asr_state.previous_pred_out[0] is previous_pred_out

    @pytest.mark.unit
    def test_get_active_speakers_info(self, asr_model, diar_model):
        """Test get_active_speakers_info with both empty and active speakers"""
        instance_manager = MultiTalkerInstanceManager(
            asr_model=asr_model,
            diar_model=diar_model,
            batch_size=2,
            max_num_of_spks=4,
            sent_break_sec=5.0,
        )
        instance_manager.reset()

        # Set up diar state with mock data
        previous_chunk_preds = torch.randn(2, 10, 4)
        instance_manager.diar_states.previous_chunk_preds = previous_chunk_preds

        # Test 1: No active speakers - should return None
        active_speakers_empty = [[], []]
        chunk_audio = torch.randn(2, 1600)
        chunk_lengths = torch.tensor([1600, 1600])

        result = instance_manager.get_active_speakers_info(active_speakers_empty, chunk_audio, chunk_lengths)
        assert result == (None, None, None, None)

        # Test 2: Active speakers - batch 0 has speaker 0, batch 1 has speakers 0 and 1
        active_speakers = [[0], [0, 1]]

        active_chunk_audio, active_chunk_lengths, active_speaker_targets, inactive_speaker_targets = (
            instance_manager.get_active_speakers_info(active_speakers, chunk_audio, chunk_lengths)
        )

        # Should have 3 active speakers total (1 from batch 0, 2 from batch 1)
        assert active_chunk_audio is not None
        assert active_chunk_audio.shape[0] == 3
        assert active_chunk_lengths.shape[0] == 3
        assert active_speaker_targets.shape[0] == 3
        assert inactive_speaker_targets.shape[0] == 3

    @pytest.mark.unit
    def test_update_seglsts(self, asr_model, diar_model):
        """Test update_seglsts method"""
        instance_manager = MultiTalkerInstanceManager(
            asr_model=asr_model,
            diar_model=diar_model,
            batch_size=2,
            max_num_of_spks=4,
            sent_break_sec=5.0,
        )
        instance_manager.reset()

        # Call update_seglsts (should not raise an error)
        offset = 0.0
        instance_manager.update_seglsts(offset=offset)

        # Verify seglsts are updated in each ASR state
        for asr_state in instance_manager.batch_asr_states:
            assert isinstance(asr_state.seglsts, list)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from contextlib import contextmanager

import torch


@contextmanager
def fp32_precision():
    """
    Workaround for precision related issues when training with bf16-true PyTorch Lightning precision setting.
    In bf16-true, PTL changes PyTorch's default dtype, which may break implicit assumptions for some models.
    This context manager restores default float32 precision and runs the computation in float32 autocast context.
    """
    default_dtype = torch.get_default_dtype()
    torch.set_default_dtype(torch.float32)
    try:
        with torch.amp.autocast(device_type="cuda" if torch.cuda.is_available() else "cpu", dtype=torch.float32):
            yield
    finally:
        torch.set_default_dtype(default_dtype)



# Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import io
import math
import operator
import random
import string
from copy import deepcopy
from typing import List
from unittest.mock import Mock, patch

import pytest
import torch
from omegaconf import DictConfig

from nemo.collections.asr.metrics.bleu import (
    BLEU,
    BLEU_TOKENIZER,
    _get_bleu_tokenizers_from_cuts,
    _move_dimension_to_the_front,
)
from nemo.collections.asr.metrics.multitask import ConstraintParser, MultiTaskMetric
from nemo.collections.asr.metrics.wer import WER, word_error_rate, word_error_rate_detail, word_error_rate_per_utt
from nemo.collections.asr.parts.submodules.ctc_decoding import (
    AbstractCTCDecoding,
    CTCBPEDecoding,
    CTCBPEDecodingConfig,
    CTCDecoding,
    CTCDecodingConfig,
)
from nemo.collections.asr.parts.submodules.multitask_decoding import AbstractMultiTaskDecoding
from nemo.collections.asr.parts.submodules.rnnt_decoding import AbstractRNNTDecoding, RNNTBPEDecoding, RNNTDecoding
from nemo.collections.asr.parts.utils.rnnt_utils import Hypothesis
from nemo.collections.common.tokenizers import CharTokenizer
from nemo.utils.config_utils import assert_dataclass_signature_match


def build_char_tokenizer_with_vocabulary(vocabulary: List[str]) -> CharTokenizer:
    with patch('pathlib.Path.open', Mock(return_value=io.StringIO('\n'.join([repr(char) for char in vocabulary])))):
        char_tokenizer = CharTokenizer('a_path_which_will_not_be_used')
    # For some reason `WERBPE` takes vocabulary size of inner tokenizer. Mock inner tokenizer.
    setattr(char_tokenizer, "tokenizer", Mock(vocab_size=char_tokenizer.vocab_size))
    return char_tokenizer


class TestWordErrorRate:

    vocabulary = [' '] + list(string.ascii_lowercase) + ["'"]
    char_tokenizer = build_char_tokenizer_with_vocabulary(vocabulary)

    def __string_to_ctc_tensor(self, txt: str, use_tokenizer: bool, as_logprobs: bool = False) -> torch.Tensor:
        # This function emulates how CTC output could like for txt
        if use_tokenizer:
            blank_id = self.char_tokenizer.vocab_size
            string_in_id_form = self.char_tokenizer.text_to_ids(txt)
        else:
            blank_id = len(self.vocabulary)
            char_to_ind = dict([(self.vocabulary[i], i) for i in range(len(self.vocabulary))])
            string_in_id_form = [char_to_ind[c] for c in txt]
        ctc_list = []
        prev_id = -1
        for c in string_in_id_form:
            # when character is repeated we need to insert CTC blank symbol
            if c != prev_id:
                ctc_list.append(c)
            else:
                ctc_list.append(blank_id)
                ctc_list.append(c)
            prev_id = c
        tensor = torch.Tensor(ctc_list).unsqueeze(0)

        if not as_logprobs:
            return tensor
        else:
            tensor = tensor.to(torch.int64)
            new_tensor = torch.nn.functional.one_hot(tensor[0], num_classes=blank_id)
            new_tensor = new_tensor.unsqueeze(0)  # [1, V, T]
            return new_tensor

    def __reference_string_to_tensor(self, txt: str, use_tokenizer: bool) -> torch.Tensor:
        # Reference tensors aren't produced by CTC logic
        if use_tokenizer:
            string_in_id_form = self.char_tokenizer.text_to_ids(txt)
        else:
            char_to_ind = dict([(self.vocabulary[i], i) for i in range(len(self.vocabulary))])
            string_in_id_form = [char_to_ind[c] for c in txt]
        return torch.Tensor(string_in_id_form).unsqueeze(0)

    def get_wer(self, wer, prediction: str, reference: str, use_tokenizer: bool):
        predictions_tensor = self.__string_to_ctc_tensor(prediction, use_tokenizer)
        targets_tensor = self.__reference_string_to_tensor(reference, use_tokenizer)
        if wer.batch_dim_index > 0:
            targets_tensor.transpose_(0, 1)
            predictions_tensor.transpose_(0, 1)
        wer(
            predictions=predictions_tensor,
            predictions_lengths=None,
            targets=targets_tensor,
            targets_lengths=torch.tensor([len(reference)]),
        )
        res, _, _ = wer.compute()
        res = res.detach().cpu()
        # return res[0] / res[1]
        return res.item()

    @pytest.mark.unit
    def test_wer_function(self):
        assert word_error_rate(hypotheses=['cat'], references=['cot']) == 1.0
        assert word_error_rate(hypotheses=['GPU'], references=['G P U']) == 1.0
        assert word_error_rate(hypotheses=['G P U'], references=['GPU']) == 3.0
        assert word_error_rate(hypotheses=['ducati motorcycle'], references=['motorcycle']) == 1.0
        assert word_error_rate(hypotheses=['ducati motorcycle'], references=['ducuti motorcycle']) == 0.5
        assert word_error_rate(hypotheses=['a B c'], references=['a b c']) == 1.0 / 3.0

        assert word_error_rate_detail(hypotheses=['cat'], references=['cot'])[0] == 1.0
        assert word_error_rate_detail(hypotheses=['GPU'], references=['G P U'])[0] == 1.0
        assert word_error_rate_detail(hypotheses=['G P U'], references=['GPU'])[0] == 3.0
        assert word_error_rate_detail(hypotheses=['ducati motorcycle'], references=['motorcycle'])[0] == 1.0
        assert word_error_rate_detail(hypotheses=['ducati motorcycle'], references=['ducuti motorcycle'])[0] == 0.5
        assert word_error_rate_detail(hypotheses=['a B c'], references=['a b c'])[0] == 1.0 / 3.0

        assert word_error_rate_detail(hypotheses=['cat'], references=['']) == (
            float("inf"),
            0,
            float("inf"),
            float("inf"),
            float("inf"),
        )
        assert word_error_rate_detail(hypotheses=['cat', ''], references=['', 'gpu']) == (
            2.0,
            1,
            1.0,
            1.0,
            0.0,
        )
        assert word_error_rate_detail(hypotheses=['cat'], references=['cot']) == (1.0, 1, 0.0, 0.0, 1.0)
        assert word_error_rate_detail(hypotheses=['G P U'], references=['GPU']) == (3.0, 1, 2.0, 0.0, 1.0)
        assert word_error_rate_detail(hypotheses=[''], references=['ducuti motorcycle'], use_cer=True) == (
            1.0,
            17,
            0.0,
            1.0,
            0.0,
        )

        assert word_error_rate_per_utt(hypotheses=['kat'], references=['cat']) == ([1.0], 1.0)
        assert word_error_rate_per_utt(hypotheses=['cat', ''], references=['', 'gpu']) == ([float("inf"), 1.0], 2.0)
        assert word_error_rate_per_utt(
            hypotheses=['ducuti motorcycle', 'G P U'], references=['ducati motorcycle', 'GPU']
        ) == ([0.5, 3.0], 4 / 3)
        assert word_error_rate_per_utt(
            hypotheses=['ducuti motorcycle', 'G P U'], references=['ducati motorcycle', 'GPU'], use_cer=True
        ) == ([1 / 17, 2 / 3], 0.15)

    @pytest.mark.unit
    @pytest.mark.parametrize("batch_dim_index", [0, 1])
    @pytest.mark.parametrize("test_wer_bpe", [False, True])
    def test_wer_metric_simple(self, batch_dim_index, test_wer_bpe):
        assert self.get_wer_ctc('cat', 'cot', test_wer_bpe) == 1.0
        assert self.get_wer_ctc('gpu', 'g p u', test_wer_bpe) == 1.0
        assert self.get_wer_ctc('g p u', 'gpu', test_wer_bpe) == 3.0
        assert self.get_wer_ctc('ducati motorcycle', 'motorcycle', test_wer_bpe) == 1.0
        assert self.get_wer_ctc('ducati motorcycle', 'ducuti motorcycle', test_wer_bpe) == 0.5
        assert abs(self.get_wer_ctc('a f c', 'a b c', test_wer_bpe) - 1.0 / 3.0) < 1e-6

    @pytest.mark.unit
    @pytest.mark.parametrize("test_wer_bpe", [False, True])
    def test_wer_metric_randomized(self, test_wer_bpe):
        """This test relies on correctness of word_error_rate function."""

        def __random_string(length):
            return ''.join(random.choice(''.join(self.vocabulary)) for _ in range(length))

        for test_id in range(256):
            n1 = random.randint(1, 512)
            n2 = random.randint(1, 512)
            s1 = __random_string(n1)
            s2 = __random_string(n2)
            # skip empty strings as reference
            if s2.strip():
                assert (
                    abs(
                        self.get_wer_ctc(prediction=s1, reference=s2, test_wer_bpe=test_wer_bpe)
                        - word_error_rate(hypotheses=[s1], references=[s2])
                    )
                    < 1e-6
                )

    @pytest.mark.unit
    @pytest.mark.parametrize("test_wer_bpe", [False, True])
    def test_wer_metric_decode(self, test_wer_bpe):
        decoding_config = {'strategy': 'greedy'}
        if test_wer_bpe:
            decoding = CTCBPEDecoding(decoding_config, self.char_tokenizer)
            wer = WER(decoding, use_cer=False)
        else:
            decoding = CTCDecoding(decoding_config, self.vocabulary.copy())
            wer = WER(decoding, use_cer=False)

        tokens = self.__string_to_ctc_tensor('cat', use_tokenizer=test_wer_bpe)[0].int().numpy().tolist()
        assert tokens == [3, 1, 20]

        tokens_decoded = wer.decoding.decode_ids_to_tokens(tokens)
        assert tokens_decoded == ['c', 'a', 't']

        str_decoded = wer.decoding.decode_ids_to_str(tokens)
        assert str_decoded == 'cat'

    @pytest.mark.unit
    @pytest.mark.parametrize("batch_dim_index", [0, 1])
    @pytest.mark.parametrize("test_wer_bpe", [False, True])
    def test_wer_metric_return_hypothesis(self, batch_dim_index, test_wer_bpe):
        decoding_config = {'strategy': 'greedy', 'batch_dim_index': batch_dim_index}
        wer = WER(CTCDecoding(decoding_config, self.vocabulary), use_cer=False)

        tensor = self.__string_to_ctc_tensor('cat', test_wer_bpe, as_logprobs=True).int()
        if batch_dim_index > 0:
            tensor.transpose_(0, 1)

        # pass batchsize 1 tensor, get back list of length 1 Hypothesis
        wer.decoding.preserve_alignments = True
        hyp = wer.decoding.ctc_decoder_predictions_tensor(tensor, return_hypotheses=True)
        hyp = hyp[0]
        assert isinstance(hyp, Hypothesis)

        sample = tensor[0] if batch_dim_index == 0 else tensor[:, 0, :]
        assert (hyp.y_sequence - torch.tensor([3, 1, 20])).sum() == 0
        assert hyp.score == 3  # sum of number of tokens in one hot representation
        assert hyp.text == 'cat'
        assert (hyp.alignments[0] == sample).all()
        assert hyp.length == 0

        length = torch.tensor([tensor.shape[1 - batch_dim_index]], dtype=torch.long)

        # pass batchsize 1 tensor, get back list of length 1 Hypothesis [add length info]
        hyp = wer.decoding.ctc_decoder_predictions_tensor(tensor, decoder_lengths=length, return_hypotheses=True)
        hyp = hyp[0]
        assert isinstance(hyp, Hypothesis)
        assert hyp.length == 3

    @pytest.mark.unit
    @pytest.mark.parametrize("batch_dim_index", [0, 1])
    @pytest.mark.parametrize("test_wer_bpe", [False, True])
    def test_wer_metric_subword_return_hypothesis(self, batch_dim_index, test_wer_bpe):
        decoding_config = {'strategy': 'greedy', 'batch_dim_index': batch_dim_index}
        wer = WER(CTCBPEDecoding(decoding_config, self.char_tokenizer), use_cer=False)

        tensor = self.__string_to_ctc_tensor('cat', test_wer_bpe, as_logprobs=True).int()
        if batch_dim_index > 0:
            tensor.transpose_(0, 1)

        # pass batchsize 1 tensor, get back list of length 1 Hypothesis
        wer.decoding.preserve_alignments = True
        hyp = wer.decoding.ctc_decoder_predictions_tensor(tensor, return_hypotheses=True)
        hyp = hyp[0]
        assert isinstance(hyp, Hypothesis)

        sample = tensor[0] if batch_dim_index == 0 else tensor[:, 0, :]
        assert (hyp.y_sequence - torch.tensor([3, 1, 20])).sum() == 0
        assert hyp.score == 3  # sum of number of tokens in one hot representation
        assert hyp.text == 'cat'
        assert (hyp.alignments[0] == sample).all()
        assert hyp.length == 0

        length = torch.tensor([tensor.shape[1 - batch_dim_index]], dtype=torch.long)

        # pass batchsize 1 tensor, get back list of length 1 Hypothesis [add length info]
        hyp = wer.decoding.ctc_decoder_predictions_tensor(tensor, decoder_lengths=length, return_hypotheses=True)
        hyp = hyp[0]
        assert isinstance(hyp, Hypothesis)
        assert hyp.length == 3

    def get_wer_ctc(self, prediction: str, reference: str, test_wer_bpe: bool):
        ctc_decoder_predictions_tensor_mock = Mock(
            return_value=[Hypothesis(score=1.0, y_sequence=[], text=prediction)]
        )
        if test_wer_bpe:
            decoding = Mock(
                blank_id=self.char_tokenizer.tokenizer.vocab_size,
                tokenizer=deepcopy(self.char_tokenizer),
                ctc_decoder_predictions_tensor=ctc_decoder_predictions_tensor_mock,
                decode_ids_to_str=self.char_tokenizer.ids_to_text,
                spec=CTCBPEDecoding,
            )
            wer = WER(decoding, use_cer=False)
        else:
            decoding = Mock(
                blank_id=len(self.vocabulary),
                labels_map=self.vocabulary.copy(),
                ctc_decoder_predictions_tensor=ctc_decoder_predictions_tensor_mock,
                decode_ids_to_str=self.decode_token_to_str_with_vocabulary_mock,
                spec=CTCDecoding,
            )
            wer = WER(decoding, use_cer=False)

        predictions_tensor = self.__string_to_ctc_tensor(prediction, test_wer_bpe)
        targets_tensor = self.__reference_string_to_tensor(reference, test_wer_bpe)

        wer(
            predictions=predictions_tensor,
            predictions_lengths=None,
            targets=targets_tensor,
            targets_lengths=torch.tensor([len(reference)]),
        )
        res, _, _ = wer.compute()
        res = res.detach().cpu()
        # return res[0] / res[1]
        return res.item()

    def decode_token_to_str_with_vocabulary_mock(self, ids):
        return ''.join([self.vocabulary[id_] for id_ in ids])

    def get_wer_rnnt(self, prediction: str, reference: str, batch_dim_index: int, test_wer_bpe: bool):
        rnnt_decoder_predictions_tensor_mock = Mock(
            return_value=[Hypothesis(score=1.0, y_sequence=[], text=prediction)]
        )
        if test_wer_bpe:
            decoding = Mock(
                blank_id=self.char_tokenizer.tokenizer.vocab_size,
                tokenizer=deepcopy(self.char_tokenizer),
                rnnt_decoder_predictions_tensor=rnnt_decoder_predictions_tensor_mock,
                decode_ids_to_str=self.char_tokenizer.ids_to_text,
                spec=RNNTBPEDecoding,
            )
            wer = WER(decoding, batch_dim_index=batch_dim_index, use_cer=False)
        else:
            decoding = Mock(
                blank_id=len(self.vocabulary),
                labels_map=self.vocabulary.copy(),
                rnnt_decoder_predictions_tensor=rnnt_decoder_predictions_tensor_mock,
                decode_ids_to_str=self.decode_token_to_str_with_vocabulary_mock,
                spec=RNNTDecoding,
            )
            wer = WER(decoding, batch_dim_index=batch_dim_index, use_cer=False)
        targets_tensor = self.__reference_string_to_tensor(reference, test_wer_bpe)
        if wer.batch_dim_index > 0:
            targets_tensor.transpose_(0, 1)

        # Create proper predictions tensor instead of passing None
        predictions_tensor = self.__string_to_ctc_tensor(prediction, test_wer_bpe)
        if wer.batch_dim_index > 0:
            predictions_tensor.transpose_(0, 1)

        wer(
            predictions=predictions_tensor,
            predictions_lengths=None,
            targets=targets_tensor,
            targets_lengths=torch.tensor([len(reference)]),
        )
        res, _, _ = wer.compute()
        res = res.detach().cpu()
        # return res[0] / res[1]
        return res.item()

    @pytest.mark.unit
    @pytest.mark.parametrize("batch_dim_index", [0, 1])
    @pytest.mark.parametrize("test_wer_bpe", [False, True])
    def test_rnnt_wer_metric_simple(self, batch_dim_index, test_wer_bpe):
        assert self.get_wer_rnnt('cat', 'cot', batch_dim_index, test_wer_bpe) == 1.0
        assert self.get_wer_rnnt('gpu', 'g p u', batch_dim_index, test_wer_bpe) == 1.0
        assert self.get_wer_rnnt('g p u', 'gpu', batch_dim_index, test_wer_bpe) == 3.0
        assert self.get_wer_rnnt('ducati motorcycle', 'motorcycle', batch_dim_index, test_wer_bpe) == 1.0
        assert self.get_wer_rnnt('ducati motorcycle', 'ducuti motorcycle', batch_dim_index, test_wer_bpe) == 0.5
        assert abs(self.get_wer_rnnt('a f c', 'a b c', batch_dim_index, test_wer_bpe) - 1.0 / 3.0) < 1e-6

    @pytest.mark.unit
    @pytest.mark.parametrize("test_wer_bpe", [False, True])
    def test_rnnt_wer_metric_randomized(self, test_wer_bpe):
        """This test relies on correctness of word_error_rate function."""

        def __random_string(length):
            return ''.join(random.choice(''.join(self.vocabulary)) for _ in range(length))

        for test_id in range(256):
            n1 = random.randint(1, 512)
            n2 = random.randint(1, 512)
            s1 = __random_string(n1)
            s2 = __random_string(n2)
            # skip empty strings as reference
            if s2.strip():
                assert (
                    abs(
                        self.get_wer_rnnt(prediction=s1, reference=s2, batch_dim_index=0, test_wer_bpe=test_wer_bpe)
                        - word_error_rate(hypotheses=[s1], references=[s2])
                    )
                    < 1e-6
                )

    @pytest.mark.unit
    def test_char_decoding_logprobs(self):
        B, T, V = 1, 8, len(self.vocabulary)
        torch.manual_seed(0)
        decoder_outputs = torch.randn(B, T, V, dtype=torch.float32)
        decoder_lens = torch.randint(0, T, size=[B], dtype=torch.int32)
        decoder_lens[torch.randint(0, B, [1])[0]] = T

        decoding_cfg = CTCDecodingConfig()
        decoding = CTCDecoding(decoding_cfg, vocabulary=self.vocabulary)

        hyp = decoding.ctc_decoder_predictions_tensor(decoder_outputs, decoder_lens, return_hypotheses=True)
        hyp = hyp[0]  # type: Hypothesis
        assert isinstance(hyp.y_sequence, torch.Tensor)
        assert hyp.length == torch.tensor(T, dtype=torch.int32)
        assert hyp.text != ''
        assert len(hyp.timestamp) == 0
        assert hyp.alignments is None

        # Preserve timestamps and alignments
        decoding_cfg = CTCDecodingConfig(preserve_alignments=True, compute_timestamps=True)
        decoding = CTCDecoding(decoding_cfg, vocabulary=self.vocabulary)

        hyp = decoding.ctc_decoder_predictions_tensor(decoder_outputs, decoder_lens, return_hypotheses=True)
        hyp = hyp[0]  # type: Hypothesis
        assert isinstance(hyp.y_sequence, torch.Tensor)
        assert hyp.length == torch.tensor(T, dtype=torch.int32)
        assert hyp.text != ''
        assert len(hyp.timestamp) == 4
        assert hyp.alignments is not None

    @pytest.mark.unit
    def test_subword_decoding_logprobs(self):
        B, T, V = 1, 8, self.char_tokenizer.vocab_size
        torch.manual_seed(0)
        decoder_outputs = torch.randn(B, T, V, dtype=torch.float32)
        decoder_lens = torch.randint(0, T, size=[B], dtype=torch.int32)
        decoder_lens[torch.randint(0, B, [1])[0]] = T

        decoding_cfg = CTCBPEDecodingConfig()
        decoding = CTCBPEDecoding(decoding_cfg, tokenizer=self.char_tokenizer)

        hyp = decoding.ctc_decoder_predictions_tensor(decoder_outputs, decoder_lens, return_hypotheses=True)
        hyp = hyp[0]  # type: Hypothesis
        assert isinstance(hyp.y_sequence, torch.Tensor)
        assert hyp.length == torch.tensor(T, dtype=torch.int32)
        assert hyp.text != ''
        assert len(hyp.timestamp) == 0
        assert hyp.alignments is None

        # Preserve timestamps and alignments
        decoding_cfg = CTCBPEDecodingConfig(preserve_alignments=True, compute_timestamps=True)
        decoding = CTCBPEDecoding(decoding_cfg, tokenizer=self.char_tokenizer)

        hyp = decoding.ctc_decoder_predictions_tensor(decoder_outputs, decoder_lens, return_hypotheses=True)
        hyp = hyp[0]  # type: Hypothesis
        assert isinstance(hyp.y_sequence, torch.Tensor)
        assert hyp.length == torch.tensor(T, dtype=torch.int32)
        assert hyp.text != ''
        assert len(hyp.timestamp) == 4
        assert hyp.alignments is not None

    @pytest.mark.unit
    def test_char_decoding_labels(self):
        B, T, V = 1, 8, len(self.vocabulary)
        torch.manual_seed(0)
        decoder_outputs = torch.randint(0, V + 1, size=[B, T], dtype=torch.float32)
        decoder_lens = torch.randint(0, T, size=[B], dtype=torch.int32)
        decoder_lens[torch.randint(0, B, [1])[0]] = T

        decoding_cfg = CTCDecodingConfig()
        decoding = CTCDecoding(decoding_cfg, vocabulary=self.vocabulary)

        hyp = decoding.ctc_decoder_predictions_tensor(decoder_outputs, decoder_lens, return_hypotheses=True)
        hyp = hyp[0]  # type: Hypothesis
        assert isinstance(hyp.y_sequence, torch.Tensor)
        assert hyp.length == torch.tensor(T, dtype=torch.int32)
        assert hyp.text != ''
        assert len(hyp.timestamp) == 0
        assert hyp.alignments is None

        # Preserve timestamps and alignments
        decoding_cfg = CTCDecodingConfig(preserve_alignments=True, compute_timestamps=True)
        decoding = CTCDecoding(decoding_cfg, vocabulary=self.vocabulary)

        # Cannot compute alignments from labels
        with pytest.raises(ValueError):
            _ = decoding.ctc_decoder_predictions_tensor(decoder_outputs, decoder_lens, return_hypotheses=True)

        # Preserve timestamps
        decoding_cfg = CTCDecodingConfig(preserve_alignments=False, compute_timestamps=True)
        decoding = CTCDecoding(decoding_cfg, vocabulary=self.vocabulary)

        hyp = decoding.ctc_decoder_predictions_tensor(decoder_outputs, decoder_lens, return_hypotheses=True)
        hyp = hyp[0]  # type: Hypothesis
        assert isinstance(hyp.y_sequence, torch.Tensor)
        assert hyp.length == torch.tensor(T, dtype=torch.int32)
        assert hyp.text != ''
        assert len(hyp.timestamp) == 4
        assert hyp.alignments is None

    @pytest.mark.unit
    def test_subword_decoding_logprobs(self):
        B, T, V = 1, 8, self.char_tokenizer.vocab_size
        torch.manual_seed(0)
        decoder_outputs = torch.randn(B, T, V, dtype=torch.float32)
        decoder_lens = torch.randint(0, T, size=[B], dtype=torch.int32)
        decoder_lens[torch.randint(0, B, [1])[0]] = T

        decoding_cfg = CTCBPEDecodingConfig()
        decoding = CTCBPEDecoding(decoding_cfg, tokenizer=self.char_tokenizer)

        hyp = decoding.ctc_decoder_predictions_tensor(decoder_outputs, decoder_lens, return_hypotheses=True)
        hyp = hyp[0]  # type: Hypothesis
        assert isinstance(hyp.y_sequence, torch.Tensor)
        assert hyp.length == torch.tensor(T, dtype=torch.int32)
        assert hyp.text != ''
        assert len(hyp.timestamp) == 0
        assert hyp.alignments is None

        # Preserve timestamps and alignments
        decoding_cfg = CTCBPEDecodingConfig(preserve_alignments=True, compute_timestamps=True)
        decoding = CTCBPEDecoding(decoding_cfg, tokenizer=self.char_tokenizer)

        hyp = decoding.ctc_decoder_predictions_tensor(decoder_outputs, decoder_lens, return_hypotheses=True)
        hyp = hyp[0]  # type: Hypothesis
        assert isinstance(hyp.y_sequence, torch.Tensor)
        assert hyp.length == torch.tensor(T, dtype=torch.int32)
        assert hyp.text != ''
        assert len(hyp.timestamp) == 4
        assert hyp.alignments is not None

    @pytest.mark.unit
    def test_subword_decoding_labels(self):
        B, T, V = 1, 8, self.char_tokenizer.vocab_size
        torch.manual_seed(0)
        decoder_outputs = torch.randint(0, V + 1, size=[B, T], dtype=torch.float32)
        decoder_lens = torch.randint(0, T, size=[B], dtype=torch.int32)
        decoder_lens[torch.randint(0, B, [1])[0]] = T

        decoding_cfg = CTCBPEDecodingConfig()
        decoding = CTCBPEDecoding(decoding_cfg, tokenizer=self.char_tokenizer)

        hyp = decoding.ctc_decoder_predictions_tensor(decoder_outputs, decoder_lens, return_hypotheses=True)
        hyp = hyp[0]  # type: Hypothesis
        assert isinstance(hyp.y_sequence, torch.Tensor)
        assert hyp.length == torch.tensor(T, dtype=torch.int32)
        assert hyp.text != ''
        assert len(hyp.timestamp) == 0
        assert hyp.alignments is None

        # Preserve timestamps and alignments
        decoding_cfg = CTCBPEDecodingConfig(preserve_alignments=True, compute_timestamps=True)
        decoding = CTCBPEDecoding(decoding_cfg, tokenizer=self.char_tokenizer)

        # Cannot compute alignments from labels
        with pytest.raises(ValueError):
            _ = decoding.ctc_decoder_predictions_tensor(decoder_outputs, decoder_lens, return_hypotheses=True)

        # Preserve timestamps
        decoding_cfg = CTCBPEDecodingConfig(preserve_alignments=False, compute_timestamps=True)
        decoding = CTCBPEDecoding(decoding_cfg, tokenizer=self.char_tokenizer)

        hyp = decoding.ctc_decoder_predictions_tensor(decoder_outputs, decoder_lens, return_hypotheses=True)
        hyp = hyp[0]  # type: Hypothesis
        assert isinstance(hyp.y_sequence, torch.Tensor)
        assert hyp.length == torch.tensor(T, dtype=torch.int32)
        assert hyp.text != ''
        assert len(hyp.timestamp) == 4
        assert hyp.alignments is None


class TestBLEUHelperFunctions:
    """Test BLEU helper functions"""

    @pytest.mark.unit
    def test_get_bleu_tokenizers_from_cuts_missing_tokenizer(self):
        """Test handling cuts without BLEU tokenizer"""
        cut = Mock()
        cut.custom = {}  # Empty dict, no BLEU_TOKENIZER key
        cut.first_non_padding_cut = cut

        cuts = [cut]
        tokenizers = _get_bleu_tokenizers_from_cuts(cuts)

        assert tokenizers == [None]

    @pytest.mark.unit
    def test_move_dimension_to_the_front(self):
        """Test moving tensor dimensions"""
        # Create test tensor [batch, time, features]
        tensor = torch.randn(2, 10, 128)

        # Move time dimension to front
        moved = _move_dimension_to_the_front(tensor, 1)
        assert moved.shape == (10, 2, 128)

        # Move features dimension to front
        moved = _move_dimension_to_the_front(tensor, 2)
        assert moved.shape == (128, 2, 10)

        # Move batch dimension to front (should be unchanged)
        moved = _move_dimension_to_the_front(tensor, 0)
        assert moved.shape == (2, 10, 128)


class TestBLEUMetric:
    """Test BLEU metric functionality"""

    vocabulary = [' '] + list(string.ascii_lowercase) + ["'"] + ['你', '好', '世', '界', '朋', '友']
    char_tokenizer = build_char_tokenizer_with_vocabulary(vocabulary)

    def create_mock_decoding(self, decode_type="ctc"):
        """Create mock decoding instance"""
        decoding = None
        if decode_type == "ctc":
            decoding = Mock(spec=AbstractCTCDecoding)
            decoding.decode_ids_to_str = lambda tokens: ''.join([self.vocabulary[id_] for id_ in tokens])
            decoding.ctc_decoder_predictions_tensor = Mock(
                return_value=[Hypothesis(score=1.0, y_sequence=[], text="hello world")]
            )
        elif decode_type == "rnnt":
            decoding = Mock(spec=AbstractRNNTDecoding)
            decoding.decode_ids_to_str = lambda tokens: ''.join([self.vocabulary[id_] for id_ in tokens])
            decoding.rnnt_decoder_predictions_tensor = Mock(
                return_value=[Hypothesis(score=1.0, y_sequence=[], text="hello world")]
            )
        elif decode_type == "multitask":
            decoding = Mock(spec=AbstractMultiTaskDecoding)
            decoding.decode_ids_to_str = lambda tokens: ''.join([self.vocabulary[id_] for id_ in tokens])
            decoding.decode_predictions_tensor = Mock(
                return_value=[Hypothesis(score=1.0, y_sequence=[], text="hello world")]
            )
        else:
            raise TypeError(f"`decode_type:` {decode_type} is invalid type for `create_mock_decoding'")
        return decoding

    def __reference_string_to_tensor(self, txt: str) -> torch.Tensor:
        """Convert reference string to tensor"""
        char_to_ind = dict([(self.vocabulary[i], i) for i in range(len(self.vocabulary))])
        string_in_id_form = [char_to_ind[c] for c in txt]
        return torch.tensor(string_in_id_form).unsqueeze(0)

    @pytest.mark.unit
    @pytest.mark.parametrize("batch_dim_index", [0, 1])
    @pytest.mark.parametrize("decode_type", ["ctc", "rnnt", "multitask"])
    def test_bleu_initialization(self, batch_dim_index, decode_type):
        """Test BLEU metric initialization with different decoders"""
        decoding = self.create_mock_decoding(decode_type)

        bleu = BLEU(
            decoding=decoding, batch_dim_index=batch_dim_index, bleu_tokenizer="13a", n_gram=4, lowercase=False
        )

        assert bleu.batch_dim_index == batch_dim_index
        assert bleu.decoding == decoding
        assert bleu.n_gram == 4
        assert bleu.decode is not None

    @pytest.mark.unit
    def test_bleu_update_basic(self):
        """Test basic BLEU update functionality"""
        decoding = self.create_mock_decoding("ctc")
        bleu = BLEU(decoding=decoding, bleu_tokenizer="13a")

        # Create test tensors
        batch_size = 2
        predictions = torch.randn(batch_size, 100, len(self.vocabulary))
        predictions_lengths = torch.tensor([80, 90])
        targets = self.__reference_string_to_tensor("hello world")
        targets = targets.repeat(batch_size, 1)
        targets_lengths = torch.tensor([11, 11])  # Length of "hello world"

        # Mock decode function to return expected predictions
        decoding.ctc_decoder_predictions_tensor.return_value = [
            Hypothesis(score=1.0, y_sequence=[], text="hello world"),
            Hypothesis(score=1.0, y_sequence=[], text="hello earth"),
        ]

        bleu.update(
            predictions=predictions,
            predictions_lengths=predictions_lengths,
            targets=targets,
            targets_lengths=targets_lengths,
        )

        # Verify decode was called
        decoding.ctc_decoder_predictions_tensor.assert_called_once()

    @pytest.mark.unit
    def test_bleu_update_empty_predictions(self):
        """Test BLEU update with empty predictions"""
        decoding = self.create_mock_decoding("ctc")
        bleu = BLEU(decoding=decoding, bleu_tokenizer="13a")

        # Create empty predictions tensor
        predictions = torch.empty(0, 100, len(self.vocabulary))
        predictions_lengths = torch.empty(0, dtype=torch.long)
        targets = torch.empty(0, 50, dtype=torch.long)
        targets_lengths = torch.empty(0, dtype=torch.long)

        # Should not raise an error
        bleu.update(
            predictions=predictions,
            predictions_lengths=predictions_lengths,
            targets=targets,
            targets_lengths=targets_lengths,
        )

        # Decode should not be called for empty predictions
        decoding.ctc_decoder_predictions_tensor.assert_not_called()

    @pytest.mark.unit
    @pytest.mark.parametrize("batch_dim_index", [0, 1])
    def test_bleu_update_different_batch_dims(self, batch_dim_index):
        """Test BLEU update with different batch dimension indices"""
        decoding = self.create_mock_decoding("ctc")
        bleu = BLEU(decoding=decoding, batch_dim_index=batch_dim_index, bleu_tokenizer="13a")

        batch_size = 2
        time_steps = 50
        vocab_size = len(self.vocabulary)

        if batch_dim_index == 0:
            predictions = torch.randn(batch_size, time_steps, vocab_size)
            targets = torch.randint(0, vocab_size, (batch_size, time_steps))
        else:
            predictions = torch.randn(time_steps, batch_size, vocab_size)
            targets = torch.randint(0, vocab_size, (time_steps, batch_size))

        predictions_lengths = torch.tensor([40, 45])
        targets_lengths = torch.tensor([35, 40])

        decoding.ctc_decoder_predictions_tensor.return_value = [
            Hypothesis(score=1.0, y_sequence=[], text="test"),
            Hypothesis(score=1.0, y_sequence=[], text="sample"),
        ]

        bleu.update(
            predictions=predictions,
            predictions_lengths=predictions_lengths,
            targets=targets,
            targets_lengths=targets_lengths,
        )

        decoding.ctc_decoder_predictions_tensor.assert_called_once()

    @pytest.mark.unit
    def test_bleu_with_cuts_tokenizers(self):
        """Test BLEU with different tokenizers from cuts"""
        decoding = self.create_mock_decoding("ctc")
        bleu = BLEU(decoding=decoding, check_cuts_for_tokenizers=True, bleu_tokenizer="13a")

        # Create cuts with different tokenizers
        cut1 = Mock()
        cut1.custom = {BLEU_TOKENIZER: "13a"}  # Use dict instead of Mock
        cut1.first_non_padding_cut = cut1

        cut2 = Mock()
        cut2.custom = {BLEU_TOKENIZER: "zh"}  # Use dict instead of Mock
        cut2.first_non_padding_cut = cut2

        cuts = [cut1, cut2]

        batch_size = 2
        predictions = torch.randn(batch_size, 100, len(self.vocabulary))
        predictions_lengths = torch.tensor([80, 90])
        targets = self.__reference_string_to_tensor("test")
        targets = targets.repeat(batch_size, 1)
        targets_lengths = torch.tensor([4, 4])

        decoding.ctc_decoder_predictions_tensor.return_value = [
            Hypothesis(score=1.0, y_sequence=[], text="test"),
            Hypothesis(score=1.0, y_sequence=[], text="测试"),
        ]

        bleu.update(
            predictions=predictions,
            predictions_lengths=predictions_lengths,
            targets=targets,
            targets_lengths=targets_lengths,
            cuts=cuts,
        )

        decoding.ctc_decoder_predictions_tensor.assert_called_once()

    @pytest.mark.unit
    def test_bleu_cuts_length_mismatch(self):
        """Test BLEU with mismatched cuts and batch size"""
        decoding = self.create_mock_decoding("ctc")
        bleu = BLEU(decoding=decoding, check_cuts_for_bleu_tokenizers=True)

        # Create cuts with wrong length
        cuts = [Mock()]  # Only 1 cut

        batch_size = 2  # But 2 samples
        predictions = torch.randn(batch_size, 100, len(self.vocabulary))
        predictions_lengths = torch.tensor([80, 90])
        targets = torch.randint(0, len(self.vocabulary), (batch_size, 50))
        targets_lengths = torch.tensor([40, 45])

        with pytest.raises(AssertionError, match="BLEU metrics configured for multiple tokenizers"):
            bleu.update(
                predictions=predictions,
                predictions_lengths=predictions_lengths,
                targets=targets,
                targets_lengths=targets_lengths,
                cuts=cuts,
            )

    @pytest.mark.unit
    def test_bleu_perfect_match(self):
        """Test BLEU calculation with perfect prediction match"""
        decoding = self.create_mock_decoding("ctc")
        bleu = BLEU(decoding=decoding, bleu_tokenizer="13a", n_gram=1)

        # Use very simple text to debug
        perfect_text = "hello"
        predictions = torch.randn(1, 100, len(self.vocabulary))
        predictions_lengths = torch.tensor([80])
        targets = self.__reference_string_to_tensor(perfect_text)
        targets_lengths = torch.tensor([targets.shape[1]]).unsqueeze(0)

        # Mock decode to return exact match
        decoding.ctc_decoder_predictions_tensor.return_value = [
            Hypothesis(score=1.0, y_sequence=[], text=perfect_text)
        ]

        bleu.update(
            predictions=predictions,
            predictions_lengths=predictions_lengths,
            targets=targets,
            targets_lengths=targets_lengths,
        )

        result = bleu.compute(return_all_metrics=False)

        # Perfect match should give BLEU score close to 1.0 (allow small tolerance)
        assert result["bleu"].item() == 1.0

    @pytest.mark.unit
    def test_bleu_no_match(self):
        """Test BLEU calculation with no matching words"""
        decoding = self.create_mock_decoding("ctc")
        bleu = BLEU(decoding=decoding, bleu_tokenizer="13a")

        predictions = torch.randn(1, 100, len(self.vocabulary))
        predictions_lengths = torch.tensor([80])
        targets = self.__reference_string_to_tensor("hello world")
        targets_lengths = torch.tensor([11])  # "hello world"

        # Mock decode to return completely different text
        decoding.ctc_decoder_predictions_tensor.return_value = [Hypothesis(score=1.0, y_sequence=[], text="cat dog")]

        bleu.update(
            predictions=predictions,
            predictions_lengths=predictions_lengths,
            targets=targets,
            targets_lengths=targets_lengths,
        )

        result = bleu.compute(return_all_metrics=False)
        # No matching words should give BLEU score of 0.0
        assert result["bleu"].item() == 0.0

    @pytest.mark.unit
    @pytest.mark.parametrize("n_gram", [1, 2, 3, 4])
    def test_bleu_partial_match(self, n_gram):
        """Test BLEU calculation with partial word matches."""
        decoding = self.create_mock_decoding("ctc")
        bleu = BLEU(decoding=decoding, bleu_tokenizer="13a", n_gram=n_gram)

        predictions = torch.randn(1, 100, len(self.vocabulary))
        predictions_lengths = torch.tensor([80])
        targets = self.__reference_string_to_tensor("the quick brown fox jumps")
        targets_lengths = torch.tensor([targets.shape[1]])  # Use actual tensor length

        # Mock decode to return partial match with overlapping n-grams
        decoding.ctc_decoder_predictions_tensor.return_value = [
            Hypothesis(score=1.0, y_sequence=[], text="the quick brown fox runs")
        ]

        bleu.update(
            predictions=predictions,
            predictions_lengths=predictions_lengths,
            targets=targets,
            targets_lengths=targets_lengths,
        )

        result = bleu.compute(return_all_metrics=False)["bleu"]

        # Manual BLEU calculation for different n-gram levels:
        # Reference: "the quick brown fox jumps" vs Prediction: "the quick brown fox runs"
        # 1-grams: 4/5 matches (the, quick, brown, fox match; jumps≠runs)
        # 2-grams: 3/4 matches (the quick, quick brown, brown fox match; fox jumps≠fox runs)
        # 3-grams: 2/3 matches (the quick brown, quick brown fox match; brown fox jumps≠brown fox runs)
        # 4-grams: 1/2 matches (the quick brown fox matches; quick brown fox jumps≠quick brown fox runs)

        p1, p2, p3, p4 = 4 / 5, 3 / 4, 2 / 3, 1 / 2

        # Calculate expected BLEU for each n-gram level
        expected_bleu = -1
        if n_gram == 1:
            expected_bleu = p1
        elif n_gram == 2:
            expected_bleu = math.sqrt(p1 * p2)
        elif n_gram == 3:
            expected_bleu = (p1 * p2 * p3) ** (1 / 3)
        elif n_gram == 4:
            expected_bleu = (p1 * p2 * p3 * p4) ** (1 / 4)
        else:
            raise ValueError(f"`n_gram` value of {n_gram} is not supported by `test_bleu_partial_match")

        # BP = 1 (same length: 5 words each)
        assert (
            abs(result.item() - expected_bleu) < 0.1
        ), f"Expected BLEU ≈ {expected_bleu:.3f}, got {result.item():.3f}"

    @pytest.mark.unit
    def test_bleu_empty_prediction(self):
        """Test BLEU calculation with empty prediction"""
        decoding = self.create_mock_decoding("ctc")
        bleu = BLEU(decoding=decoding, bleu_tokenizer="13a")

        predictions = torch.randn(1, 100, len(self.vocabulary))
        predictions_lengths = torch.tensor([80])
        targets = self.__reference_string_to_tensor("hello world")
        targets_lengths = torch.tensor([11])

        # Mock decode to return empty text
        decoding.ctc_decoder_predictions_tensor.return_value = [Hypothesis(score=1.0, y_sequence=[], text="")]

        bleu.update(
            predictions=predictions,
            predictions_lengths=predictions_lengths,
            targets=targets,
            targets_lengths=targets_lengths,
        )

        result = bleu.compute()
        # Empty prediction should give BLEU score of 0.0
        assert result["bleu"].item() == 0.0

    @pytest.mark.unit
    def test_bleu_empty_reference(self):
        """Test BLEU calculation with empty reference"""
        decoding = self.create_mock_decoding("ctc")
        bleu = BLEU(decoding=decoding, bleu_tokenizer="13a")

        predictions = torch.randn(1, 100, len(self.vocabulary))
        predictions_lengths = torch.tensor([80])
        targets = self.__reference_string_to_tensor("")
        targets_lengths = torch.tensor([0])

        # Mock decode to return some text
        decoding.ctc_decoder_predictions_tensor.return_value = [
            Hypothesis(score=1.0, y_sequence=[], text="hello world")
        ]

        bleu.update(
            predictions=predictions,
            predictions_lengths=predictions_lengths,
            targets=targets,
            targets_lengths=targets_lengths,
        )

        result = bleu.compute()
        # Empty reference should give BLEU score of 0.0
        assert result["bleu"].item() == 0.0

    @pytest.mark.unit
    def test_bleu_multiple_samples(self):
        """Test BLEU calculation with multiple samples in batch"""
        decoding = self.create_mock_decoding("ctc")
        bleu = BLEU(decoding=decoding, bleu_tokenizer="13a", n_gram=2)  # Use 2-gram for exact calculation

        # Test with 3 samples: perfect match, partial match, no match
        batch_size = 3
        predictions = torch.randn(batch_size, 100, len(self.vocabulary))
        predictions_lengths = torch.tensor([80, 85, 90])

        # Create targets for each sample
        target1 = self.__reference_string_to_tensor("hello world")
        target2 = self.__reference_string_to_tensor("test case")
        target3 = self.__reference_string_to_tensor("example text")

        # Pad targets to same length
        max_len = max(target1.shape[1], target2.shape[1], target3.shape[1])
        targets = torch.zeros(batch_size, max_len, dtype=torch.long)
        targets[0, : target1.shape[1]] = target1[0]
        targets[1, : target2.shape[1]] = target2[0]
        targets[2, : target3.shape[1]] = target3[0]

        targets_lengths = torch.tensor([target1.shape[1], target2.shape[1], target3.shape[1]])

        # Mock decode with different quality predictions
        decoding.ctc_decoder_predictions_tensor.return_value = [
            Hypothesis(score=1.0, y_sequence=[], text="hello world"),  # Perfect match
            Hypothesis(score=1.0, y_sequence=[], text="test different"),  # Partial match (1/2 words)
            Hypothesis(score=1.0, y_sequence=[], text="completely wrong"),  # No match (0/2 words)
        ]

        bleu.update(
            predictions=predictions,
            predictions_lengths=predictions_lengths,
            targets=targets,
            targets_lengths=targets_lengths,
        )

        result = bleu.compute(return_all_metrics=False)["bleu"]

        # Corpus-level BLEU calculation:
        # Sample 1: "hello world" vs "hello world"
        #   1-grams: 2/2 matches, 2-grams: 1/1 matches
        # Sample 2: "test different" vs "test case"
        #   1-grams: 1/2 matches (only "test"), 2-grams: 0/1 matches
        # Sample 3: "completely wrong" vs "example text"
        #   1-grams: 0/2 matches, 2-grams: 0/1 matches
        #
        # Corpus totals:
        # p1 = (2 + 1 + 0) / (2 + 2 + 2) = 3/6 = 0.5
        # p2 = (1 + 0 + 0) / (1 + 1 + 1) = 1/3 ≈ 0.3333
        # BP = 1.0 (prediction length = reference length = 6 words total)
        # BLEU = sqrt(p1 * p2) = sqrt(0.5 * 0.3333) = sqrt(0.1667) ≈ 0.408

        p1 = 3 / 6  # 0.5
        p2 = 1 / 3  # 0.3333
        expected_bleu = math.sqrt(p1 * p2)  # ≈ 0.408

        # Allow reasonable tolerance for BLEU implementation differences
        assert (
            abs(result.item() - expected_bleu) < 0.1
        ), f"Expected BLEU ≈ {expected_bleu:.2f}, got {result.item():.2f}"

    @pytest.mark.unit
    @pytest.mark.parametrize("n_gram", [1, 2, 3, 4])
    def test_bleu_different_ngram_calculations(self, n_gram):
        """Test that different n-gram settings produce different BLEU scores"""
        decoding = self.create_mock_decoding("ctc")
        bleu = BLEU(decoding=decoding, bleu_tokenizer="13a", n_gram=n_gram)

        predictions = torch.randn(1, 100, len(self.vocabulary))
        predictions_lengths = torch.tensor([80])
        targets = self.__reference_string_to_tensor("the quick brown fox jumps")
        targets_lengths = torch.tensor([targets.shape[1]])  # Use actual tensor length

        # Mock decode with slightly different word order
        decoding.ctc_decoder_predictions_tensor.return_value = [
            Hypothesis(score=1.0, y_sequence=[], text="quick brown fox the jumps")
        ]

        bleu.update(
            predictions=predictions,
            predictions_lengths=predictions_lengths,
            targets=targets,
            targets_lengths=targets_lengths,
        )

        result = bleu.compute(return_all_metrics=False)["bleu"]
        # Different n-gram values should produce valid BLEU scores
        assert 0.0 <= result.item() <= 1.0

        # For this specific case:
        # - 1-gram should have high score (all words match)
        # - Higher n-grams should have lower scores due to word order
        if n_gram == 1:
            assert result.item() > 0.5  # Should be high since all words match
        else:
            # Higher n-grams might be lower due to order differences
            assert result.item() >= 0.0

    @pytest.mark.unit
    def test_bleu_multi_tokenization(self):
        """Test BLEU calculation with multiple tokenizers for different languages"""
        # Test without multi-tokenization (single tokenizer)
        decoding_single = self.create_mock_decoding("ctc")
        bleu_single = BLEU(
            decoding=decoding_single, bleu_tokenizer="13a", check_cuts_for_bleu_tokenizers=False, n_gram=2
        )

        # Test with multi-tokenization (tokenizers from cuts)
        decoding_multi = self.create_mock_decoding("ctc")
        bleu_multi = BLEU(decoding=decoding_multi, bleu_tokenizer="13a", check_cuts_for_bleu_tokenizers=True, n_gram=2)

        batch_size = 2
        predictions = torch.randn(batch_size, 100, len(self.vocabulary))
        predictions_lengths = torch.tensor([80, 85])

        # Create English and Chinese targets
        english_target = self.__reference_string_to_tensor("hello world")
        chinese_target = self.__reference_string_to_tensor("你好世界")  # "hello world" in Chinese

        # Pad targets to same length
        max_len = max(english_target.shape[1], chinese_target.shape[1])
        targets = torch.zeros(batch_size, max_len, dtype=torch.long)
        targets[0, : english_target.shape[1]] = english_target[0]
        targets[1, : chinese_target.shape[1]] = chinese_target[0]

        targets_lengths = torch.tensor([english_target.shape[1], chinese_target.shape[1]])

        # Create cuts specifying different tokenizers
        cut1 = Mock()
        cut1.custom = {BLEU_TOKENIZER: "13a"}  # Use dict instead of Mock
        cut1.first_non_padding_cut = cut1

        cut2 = Mock()
        cut2.custom = {BLEU_TOKENIZER: "zh"}  # Use dict instead of Mock
        cut2.first_non_padding_cut = cut2

        cuts = [cut1, cut2]

        # Mock predictions - partial matches for both languages
        english_prediction = "hello earth"  # Partial match with English reference
        chinese_prediction = "你好朋友"  # Partial match with Chinese reference

        decoding_single.ctc_decoder_predictions_tensor.return_value = [
            Hypothesis(score=1.0, y_sequence=[], text=english_prediction),
            Hypothesis(score=1.0, y_sequence=[], text=chinese_prediction),
        ]

        decoding_multi.ctc_decoder_predictions_tensor.return_value = [
            Hypothesis(score=1.0, y_sequence=[], text=english_prediction),
            Hypothesis(score=1.0, y_sequence=[], text=chinese_prediction),
        ]

        # Test single tokenizer (13a for both samples)
        bleu_single.update(
            predictions=predictions,
            predictions_lengths=predictions_lengths,
            targets=targets,
            targets_lengths=targets_lengths,
        )
        result_single = bleu_single.compute(return_all_metrics=False)["bleu"]

        # Test multi tokenizer (13a for English, zh for Chinese)
        bleu_multi.update(
            predictions=predictions,
            predictions_lengths=predictions_lengths,
            targets=targets,
            targets_lengths=targets_lengths,
            cuts=cuts,
        )
        result_multi = bleu_multi.compute(return_all_metrics=False)["bleu"]

        # Calculate expected BLEU for multi-tokenization case:
        # English (13a tokenizer): "hello earth" vs "hello world"
        #   - 1-grams: "hello" matches, "earth" ≠ "world" → 1/2 = 0.5
        #   - 2-grams: "hello earth" ≠ "hello world" → 0/1 = 0.0
        # Chinese (zh tokenizer): "你好朋友" vs "你好世界"
        #   - If character-level: 1-grams: "你","好" match, "朋","友" ≠ "世","界" → 2/4 = 0.5
        #   - If character-level: 2-grams: "你好" matches, others don't → 1/3 ≈ 0.33
        #
        # Corpus-level calculation (character-level assumption):
        # p1 = (1 + 2) / (2 + 4) = 3/6 = 0.5
        # p2 = (0 + 1) / (1 + 3) = 1/4 = 0.25
        # BLEU = sqrt(p1 * p2) = sqrt(0.5 * 0.25) = sqrt(0.125) ≈ 0.354
        expected_multi_bleu = math.sqrt(0.5 * 0.25)  # ≈ 0.354

        # Assert the two methods produce different results
        assert (
            abs(result_single.item() - result_multi.item()) > 0.01
        ), "Multi-tokenization should produce different BLEU scores"

        # Assert multi-tokenization result is within expected range
        assert (
            abs(result_multi.item() - expected_multi_bleu) < 0.1
        ), f"Multi-tokenization BLEU should be near {expected_multi_bleu:.4f}, got {result_multi.item():.4f}"

    @pytest.mark.unit
    def test_bleu_empty_prediction(self):
        """Test BLEU calculation with empty prediction"""
        decoding = self.create_mock_decoding("ctc")
        bleu = BLEU(decoding=decoding, bleu_tokenizer="13a")

        predictions = torch.randn(1, 100, len(self.vocabulary))
        predictions_lengths = torch.tensor([80])
        targets = self.__reference_string_to_tensor("hello world")
        targets_lengths = torch.tensor([11])

        # Mock decode to return empty text
        decoding.ctc_decoder_predictions_tensor.return_value = [Hypothesis(score=1.0, y_sequence=[], text="")]

        bleu.update(
            predictions=predictions,
            predictions_lengths=predictions_lengths,
            targets=targets,
            targets_lengths=targets_lengths,
        )

        result = bleu.compute()
        # Empty prediction should give BLEU score of 0.0
        assert result["bleu"].item() == 0.0

    @pytest.mark.unit
    def test_bleu_empty_reference(self):
        """Test BLEU calculation with empty reference"""
        decoding = self.create_mock_decoding("ctc")
        bleu = BLEU(decoding=decoding, bleu_tokenizer="13a")

        predictions = torch.randn(1, 100, len(self.vocabulary))
        predictions_lengths = torch.tensor([80])
        targets = self.__reference_string_to_tensor("")
        targets_lengths = torch.tensor([0])

        # Mock decode to return some text
        decoding.ctc_decoder_predictions_tensor.return_value = [
            Hypothesis(score=1.0, y_sequence=[], text="hello world")
        ]

        bleu.update(
            predictions=predictions,
            predictions_lengths=predictions_lengths,
            targets=targets,
            targets_lengths=targets_lengths,
        )

        result = bleu.compute()
        # Empty reference should give BLEU score of 0.0
        assert result["bleu"].item() == 0.0

    @pytest.mark.unit
    def test_bleu_multiple_samples(self):
        """Test BLEU calculation with multiple samples in batch"""
        decoding = self.create_mock_decoding("ctc")
        bleu = BLEU(decoding=decoding, bleu_tokenizer="13a", n_gram=2)  # Use 2-gram for exact calculation

        # Test with 3 samples: perfect match, partial match, no match
        batch_size = 3
        predictions = torch.randn(batch_size, 100, len(self.vocabulary))
        predictions_lengths = torch.tensor([80, 85, 90])

        # Create targets for each sample
        target1 = self.__reference_string_to_tensor("hello world")
        target2 = self.__reference_string_to_tensor("test case")
        target3 = self.__reference_string_to_tensor("example text")

        # Pad targets to same length
        max_len = max(target1.shape[1], target2.shape[1], target3.shape[1])
        targets = torch.zeros(batch_size, max_len, dtype=torch.long)
        targets[0, : target1.shape[1]] = target1[0]
        targets[1, : target2.shape[1]] = target2[0]
        targets[2, : target3.shape[1]] = target3[0]

        targets_lengths = torch.tensor([target1.shape[1], target2.shape[1], target3.shape[1]])

        # Mock decode with different quality predictions
        decoding.ctc_decoder_predictions_tensor.return_value = [
            Hypothesis(score=1.0, y_sequence=[], text="hello world"),  # Perfect match
            Hypothesis(score=1.0, y_sequence=[], text="test different"),  # Partial match (1/2 words)
            Hypothesis(score=1.0, y_sequence=[], text="completely wrong"),  # No match (0/2 words)
        ]

        bleu.update(
            predictions=predictions,
            predictions_lengths=predictions_lengths,
            targets=targets,
            targets_lengths=targets_lengths,
        )

        result = bleu.compute(return_all_metrics=False)["bleu"]

        # Corpus-level BLEU calculation:
        # Sample 1: "hello world" vs "hello world"
        #   1-grams: 2/2 matches, 2-grams: 1/1 matches
        # Sample 2: "test different" vs "test case"
        #   1-grams: 1/2 matches (only "test"), 2-grams: 0/1 matches
        # Sample 3: "completely wrong" vs "example text"
        #   1-grams: 0/2 matches, 2-grams: 0/1 matches
        #
        # Corpus totals:
        # p1 = (2 + 1 + 0) / (2 + 2 + 2) = 3/6 = 0.5
        # p2 = (1 + 0 + 0) / (1 + 1 + 1) = 1/3 ≈ 0.3333
        # BP = 1.0 (prediction length = reference length = 6 words total)
        # BLEU = sqrt(p1 * p2) = sqrt(0.5 * 0.3333) = sqrt(0.1667) ≈ 0.408

        p1 = 3 / 6  # 0.5
        p2 = 1 / 3  # 0.3333
        expected_bleu = math.sqrt(p1 * p2)  # ≈ 0.408

        # Allow reasonable tolerance for BLEU implementation differences
        assert (
            abs(result.item() - expected_bleu) < 0.1
        ), f"Expected BLEU ≈ {expected_bleu:.2f}, got {result.item():.2f}"

    @pytest.mark.unit
    @pytest.mark.parametrize("n_gram", [1, 2, 3, 4])
    def test_bleu_different_ngram_calculations(self, n_gram):
        """Test that different n-gram settings produce different BLEU scores"""
        decoding = self.create_mock_decoding("ctc")
        bleu = BLEU(decoding=decoding, bleu_tokenizer="13a", n_gram=n_gram)

        predictions = torch.randn(1, 100, len(self.vocabulary))
        predictions_lengths = torch.tensor([80])
        targets = self.__reference_string_to_tensor("the quick brown fox jumps")
        targets_lengths = torch.tensor([targets.shape[1]])  # Use actual tensor length

        # Mock decode with slightly different word order
        decoding.ctc_decoder_predictions_tensor.return_value = [
            Hypothesis(score=1.0, y_sequence=[], text="quick brown fox the jumps")
        ]

        bleu.update(
            predictions=predictions,
            predictions_lengths=predictions_lengths,
            targets=targets,
            targets_lengths=targets_lengths,
        )

        result = bleu.compute(return_all_metrics=False)["bleu"]
        # Different n-gram values should produce valid BLEU scores
        assert 0.0 <= result.item() <= 1.0

        # For this specific case:
        # - 1-gram should have high score (all words match)
        # - Higher n-grams should have lower scores due to word order
        if n_gram == 1:
            assert result.item() > 0.5  # Should be high since all words match
        else:
            # Higher n-grams might be lower due to order differences
            assert result.item() >= 0.0


class TestBLEUEdgeCases:
    """Test BLEU edge cases and error conditions"""

    vocabulary = [' '] + list(string.ascii_lowercase) + ["'"] + ['你', '好', '世', '界', '朋', '友']

    def create_mock_decoding(self):
        """Create minimal mock decoding"""
        decoding = Mock(spec=AbstractCTCDecoding)
        decoding.decode_ids_to_str = lambda tokens: ''.join([self.vocabulary[id_] for id_ in tokens])
        decoding.ctc_decoder_predictions_tensor = Mock(return_value=[])
        return decoding

    @pytest.mark.unit
    def test_bleu_empty_hypotheses(self):
        """Test BLEU with empty hypotheses"""
        decoding = self.create_mock_decoding()
        # Return empty text hypothesis to match the number of targets
        decoding.ctc_decoder_predictions_tensor.return_value = [
            Hypothesis(score=1.0, y_sequence=[], text="")  # Empty text instead of empty list
        ]

        bleu = BLEU(decoding=decoding)

        predictions = torch.randn(1, 100, len(self.vocabulary))
        predictions_lengths = torch.tensor([80])
        targets = torch.randint(0, len(self.vocabulary), (1, 50))
        targets_lengths = torch.tensor([40])

        # Should not raise an error even with empty hypotheses
        # The update method should handle empty hypotheses gracefully
        bleu.update(
            predictions=predictions,
            predictions_lengths=predictions_lengths,
            targets=targets,
            targets_lengths=targets_lengths,
        )

        # Compute should also handle empty state
        result = bleu.compute(return_all_metrics=False)
        assert result["bleu"].item() == 0.0  # Empty hypotheses should give BLEU score of 0.0

    @pytest.mark.unit
    def test_bleu_zero_length_targets(self):
        """Test BLEU with zero-length targets"""
        decoding = self.create_mock_decoding()
        bleu = BLEU(decoding=decoding)

        predictions = torch.randn(1, 100, len(self.vocabulary))
        predictions_lengths = torch.tensor([80])
        targets = torch.randint(0, len(self.vocabulary), (1, 50))
        targets_lengths = torch.tensor([0])  # Zero length

        decoding.ctc_decoder_predictions_tensor.return_value = [Hypothesis(score=1.0, y_sequence=[], text="test")]

        # Should handle zero-length targets gracefully
        bleu.update(
            predictions=predictions,
            predictions_lengths=predictions_lengths,
            targets=targets,
            targets_lengths=targets_lengths,
        )


class TestMultiTaskMetricConstraintFunctions:
    """Test the constraint parsing and evaluation functions"""

    def setUp(self):
        """Set up test fixtures"""
        self.parser = ConstraintParser()

    @pytest.mark.unit
    def test_static_constraint_equality(self):
        """Test static constraint with equality operator"""
        parser = ConstraintParser()
        properties = {"task": "transcribe", "lang": "en"}

        # Test successful match
        result = parser._static_constraint(operator.eq, "task", "transcribe", properties)
        assert result is True

        # Test failed match
        result = parser._static_constraint(operator.eq, "task", "translate", properties)
        assert result is False

        # Test missing key
        result = parser._static_constraint(operator.eq, "missing", "value", properties)
        assert result is False

    @pytest.mark.unit
    def test_static_constraint_inequality(self):
        """Test static constraint with inequality operator"""
        parser = ConstraintParser()
        properties = {"task": "transcribe", "lang": "en"}

        result = parser._static_constraint(operator.ne, "task", "translate", properties)
        assert result is True

        result = parser._static_constraint(operator.ne, "task", "transcribe", properties)
        assert result is False

    @pytest.mark.unit
    def test_compare_constraint(self):
        """Test comparing two properties"""
        parser = ConstraintParser()
        properties = {"source_lang": "en", "target_lang": "en", "other": "different"}

        # Test equal properties
        result = parser._compare_constraint(operator.eq, "source_lang", "target_lang", properties)
        assert result is True

        # Test unequal properties
        result = parser._compare_constraint(operator.eq, "source_lang", "other", properties)
        assert result is False

        # Test missing property
        result = parser._compare_constraint(operator.eq, "source_lang", "missing", properties)
        assert result is False

    @pytest.mark.unit
    def test_logical_operations(self):
        """Test logical AND, OR, NOT operations"""
        parser = ConstraintParser()
        properties = {"task": "transcribe", "lang": "en"}

        # Create simple constraint functions
        true_constraint = lambda p: p.get("task") == "transcribe"
        false_constraint = lambda p: p.get("task") == "translate"

        # Test AND
        result = parser._logical_and(true_constraint, false_constraint, properties)
        assert result is False

        # Test OR
        result = parser._logical_or(true_constraint, false_constraint, properties)
        assert result is True

        result = parser._logical_or(false_constraint, false_constraint, properties)
        assert result is False

        # Test NOT
        result = parser._logical_not(true_constraint, properties)
        assert result is False

        result = parser._logical_not(false_constraint, properties)
        assert result is True


class TestMultiTaskMetricConstraintParsing:
    """Test the constraint string parsing functionality"""

    def setUp(self):
        """Set up test fixtures"""
        self.parser = ConstraintParser()

    @pytest.mark.unit
    def test_simple_equality_constraint(self):
        """Test parsing simple equality constraints"""
        parser = ConstraintParser()
        constraint_fn = parser.parse_constraint(".task==transcribe")

        properties = {"task": "transcribe"}
        assert constraint_fn(properties) is True

        properties = {"task": "translate"}
        assert constraint_fn(properties) is False

    @pytest.mark.unit
    def test_simple_inequality_constraint(self):
        """Test parsing simple inequality constraints"""
        parser = ConstraintParser()
        constraint_fn = parser.parse_constraint(".task!=translate")

        properties = {"task": "transcribe"}
        assert constraint_fn(properties) is True

        properties = {"task": "translate"}
        assert constraint_fn(properties) is False

    @pytest.mark.unit
    def test_property_comparison_constraint(self):
        """Test parsing property-to-property comparisons"""
        parser = ConstraintParser()
        constraint_fn = parser.parse_constraint(".source_lang==.target_lang")

        properties = {"source_lang": "en", "target_lang": "en"}
        assert constraint_fn(properties) is True

        properties = {"source_lang": "en", "target_lang": "de"}
        assert constraint_fn(properties) is False

    @pytest.mark.unit
    def test_and_constraint(self):
        """Test parsing AND constraints"""
        parser = ConstraintParser()
        constraint_fn = parser.parse_constraint(".task==transcribe and .lang==en")

        properties = {"task": "transcribe", "lang": "en"}
        assert constraint_fn(properties) is True

        properties = {"task": "transcribe", "lang": "de"}
        assert constraint_fn(properties) is False

        properties = {"task": "translate", "lang": "en"}
        assert constraint_fn(properties) is False

    @pytest.mark.unit
    def test_or_constraint(self):
        """Test parsing OR constraints"""
        parser = ConstraintParser()
        constraint_fn = parser.parse_constraint(".task==transcribe or .task==translate")

        properties = {"task": "transcribe"}
        assert constraint_fn(properties) is True

        properties = {"task": "translate"}
        assert constraint_fn(properties) is True

        properties = {"task": "other"}
        assert constraint_fn(properties) is False

    @pytest.mark.unit
    def test_not_constraint(self):
        """Test parsing NOT constraints"""
        parser = ConstraintParser()
        constraint_fn = parser.parse_constraint("not .task==translate")

        properties = {"task": "transcribe"}
        assert constraint_fn(properties) is True

        properties = {"task": "translate"}
        assert constraint_fn(properties) is False

    @pytest.mark.unit
    def test_complex_constraint(self):
        """Test parsing complex nested constraints"""
        parser = ConstraintParser()
        constraint_fn = parser.parse_constraint(".task==transcribe and .source_lang==.target_lang")

        properties = {"task": "transcribe", "source_lang": "en", "target_lang": "en"}
        assert constraint_fn(properties) is True

        properties = {"task": "transcribe", "source_lang": "en", "target_lang": "de"}
        assert constraint_fn(properties) is False

        properties = {"task": "translate", "source_lang": "en", "target_lang": "en"}
        assert constraint_fn(properties) is False

    @pytest.mark.unit
    def test_parentheses_constraint(self):
        """Test parsing constraints with parentheses"""
        parser = ConstraintParser()
        # Basic parentheses
        constraint_fn = parser.parse_constraint("(.task==transcribe or .task==translate) and .lang==en")

        properties = {"task": "transcribe", "lang": "en"}
        assert constraint_fn(properties) is True

        properties = {"task": "translate", "lang": "en"}
        assert constraint_fn(properties) is True

        properties = {"task": "transcribe", "lang": "de"}
        assert constraint_fn(properties) is False

        properties = {"task": "other", "lang": "en"}
        assert constraint_fn(properties) is False

    @pytest.mark.unit
    def test_nested_parentheses_constraint(self):
        """Test parsing constraints with nested parentheses"""
        parser = ConstraintParser()
        constraint_fn = parser.parse_constraint("(.task==transcribe and (.lang==en or .lang==de)) or .task==translate")

        properties = {"task": "transcribe", "lang": "en"}
        assert constraint_fn(properties) is True

        properties = {"task": "transcribe", "lang": "de"}
        assert constraint_fn(properties) is True

        properties = {"task": "transcribe", "lang": "fr"}
        assert constraint_fn(properties) is False

        properties = {"task": "translate", "lang": "fr"}
        assert constraint_fn(properties) is True

    @pytest.mark.unit
    def test_parentheses_with_not_constraint(self):
        """Test parsing constraints with parentheses and NOT operator"""
        parser = ConstraintParser()
        constraint_fn = parser.parse_constraint("not (.task==transcribe and .lang==en)")

        properties = {"task": "transcribe", "lang": "en"}
        assert constraint_fn(properties) is False

        properties = {"task": "transcribe", "lang": "de"}
        assert constraint_fn(properties) is True

        properties = {"task": "translate", "lang": "en"}
        assert constraint_fn(properties) is True

    @pytest.mark.unit
    def test_complex_parentheses_constraint(self):
        """Test parsing complex constraints with multiple parentheses"""
        parser = ConstraintParser()
        constraint_fn = parser.parse_constraint(
            "(.task==transcribe or .task==translate) and (.source_lang!=.target_lang or .domain==special)"
        )

        properties = {"task": "transcribe", "source_lang": "en", "target_lang": "de", "domain": "general"}
        assert constraint_fn(properties) is True

        properties = {"task": "translate", "source_lang": "en", "target_lang": "en", "domain": "special"}
        assert constraint_fn(properties) is True

        properties = {"task": "transcribe", "source_lang": "en", "target_lang": "en", "domain": "general"}
        assert constraint_fn(properties) is False

        properties = {"task": "other", "source_lang": "en", "target_lang": "de", "domain": "general"}
        assert constraint_fn(properties) is False

    @pytest.mark.unit
    def test_invalid_constraint(self):
        """Test that invalid constraints raise errors"""
        parser = ConstraintParser()
        with pytest.raises(SyntaxError):
            parser.parse_constraint("invalid constraint format")


class TestMultiTaskMetricCutSplitting:
    """Test the cut splitting functionality"""

    def create_mock_cut(self, custom_data):
        """Helper to create mock cuts with custom data"""
        cut = Mock()
        cut.custom = custom_data
        # Handle MixedCut case
        cut.first_non_padding_cut = cut
        return cut

    @pytest.mark.unit
    def test_split_cuts_simple(self):
        """Test basic cut splitting"""
        # Create mock cuts
        cuts = [
            self.create_mock_cut({"task": "transcribe", "lang": "en"}),
            self.create_mock_cut({"task": "translate", "lang": "en"}),
            self.create_mock_cut({"task": "transcribe", "lang": "de"}),
        ]

        # Create MultiTaskMetric instance with mock metrics
        mock_model = Mock()
        mock_model.decoding = Mock()

        cfg = DictConfig(
            {
                "metrics": {
                    "wer": {
                        "constraint": ".task==transcribe",
                        "_target_": "nemo.collections.asr.metrics.WER",
                    },
                    "bleu": {
                        "constraint": ".task==translate",
                        "_target_": "nemo.collections.asr.metrics.BLEU",
                    },
                }
            }
        )

        with patch('nemo.collections.asr.metrics.multitask.MultiTaskMetric.from_config_dict') as mock_from_config:
            mock_wer = Mock()
            mock_bleu = Mock()
            mock_from_config.side_effect = [mock_wer, mock_bleu]

            multitask_metric = MultiTaskMetric(mock_model, cfg)

            cuts_split, idx_split = multitask_metric._split_cuts(cuts)

            # Check WER metric gets transcribe cuts (indices 0, 2)
            assert idx_split["wer"] == [0, 2]

            # Check BLEU metric gets translate cuts (index 1)
            assert idx_split["bleu"] == [1]

    @pytest.mark.unit
    def test_split_cuts_no_matches(self):
        """Test cut splitting with no matching cuts"""
        cuts = [
            self.create_mock_cut({"task": "other", "lang": "en"}),
        ]

        mock_model = Mock()
        mock_model.decoding = Mock()

        cfg = DictConfig(
            {
                "metrics": {
                    "wer": {
                        "constraint": ".task==transcribe",
                        "_target_": "nemo.collections.asr.metrics.WER",
                    }
                }
            }
        )

        with patch('nemo.collections.asr.metrics.multitask.MultiTaskMetric.from_config_dict') as mock_from_config:
            mock_wer = Mock()
            mock_from_config.return_value = mock_wer

            multitask_metric = MultiTaskMetric(mock_model, cfg)

            cuts_split, idx_split = multitask_metric._split_cuts(cuts)

            # No cuts should match
            assert idx_split["wer"] == []

    @pytest.mark.unit
    def test_split_cuts_empty_input(self):
        """Test cut splitting with empty input"""
        cuts = []

        mock_model = Mock()
        mock_model.decoding = Mock()

        cfg = DictConfig(
            {
                "metrics": {
                    "wer": {
                        "constraint": ".task==transcribe",
                        "_target_": "nemo.collections.asr.metrics.WER",
                    }
                }
            }
        )

        with patch('nemo.collections.asr.metrics.multitask.MultiTaskMetric.from_config_dict') as mock_from_config:
            mock_wer = Mock()
            mock_from_config.return_value = mock_wer

            multitask_metric = MultiTaskMetric(mock_model, cfg)

            cuts_split, idx_split = multitask_metric._split_cuts(cuts)

            assert idx_split["wer"] == []


class TestMultiTaskMetricUpdate:
    """Test the metric update functionality"""

    @pytest.fixture
    def mock_multitask_metric(self):
        """Create a MultiTaskMetric with mocked dependencies"""
        mock_model = Mock()
        mock_model.decoding = Mock()

        cfg = DictConfig(
            {
                "metrics": {
                    "wer": {
                        "constraint": ".task==transcribe",
                        "_target_": "nemo.collections.asr.metrics.WER",
                    },
                    "bleu": {
                        "constraint": ".task==translate",
                        "_target_": "nemo.collections.asr.metrics.BLEU",
                    },
                }
            }
        )

        with patch('nemo.collections.asr.metrics.multitask.MultiTaskMetric.from_config_dict') as mock_from_config:
            mock_wer = Mock()
            mock_bleu = Mock()
            mock_from_config.side_effect = [mock_wer, mock_bleu]

            multitask_metric = MultiTaskMetric(mock_model, cfg)

            # Store references for testing
            multitask_metric._mock_wer = mock_wer
            multitask_metric._mock_bleu = mock_bleu

            return multitask_metric

    @pytest.mark.unit
    def test_update_with_matching_cuts(self, mock_multitask_metric):
        """Test update with cuts that match constraints"""
        # Create mock cuts
        cuts = [Mock(), Mock()]
        cuts[0].custom = {"task": "transcribe"}
        cuts[0].first_non_padding_cut = cuts[0]
        cuts[1].custom = {"task": "translate"}
        cuts[1].first_non_padding_cut = cuts[1]

        # Create mock tensors
        batch_size = 2
        predictions = torch.randn(batch_size, 100, 1024)
        predictions_lengths = torch.tensor([80, 90])
        predictions_mask = torch.ones(batch_size, 100, dtype=torch.bool)
        targets = torch.randint(0, 1000, (batch_size, 50))
        targets_lengths = torch.tensor([40, 45])
        input_ids = torch.randint(0, 1000, (batch_size, 60))

        mock_multitask_metric.update(
            predictions=predictions,
            predictions_lengths=predictions_lengths,
            predictions_mask=predictions_mask,
            targets=targets,
            targets_lengths=targets_lengths,
            input_ids=input_ids,
            cuts=cuts,
        )

        # Verify WER metric was called with transcribe cut (index 0)
        mock_multitask_metric._mock_wer.update.assert_called_once()
        wer_call_args = mock_multitask_metric._mock_wer.update.call_args
        assert wer_call_args.kwargs['predictions'].shape[0] == 1  # One transcribe cut

        # Verify BLEU metric was called with translate cut (index 1)
        mock_multitask_metric._mock_bleu.update.assert_called_once()
        bleu_call_args = mock_multitask_metric._mock_bleu.update.call_args
        assert bleu_call_args.kwargs['predictions'].shape[0] == 1  # One translate cut

    @pytest.mark.unit
    def test_update_with_empty_indices(self, mock_multitask_metric):
        """Test update when no cuts match a metric's constraints"""
        # Create cuts that don't match WER constraint
        cuts = [Mock()]
        cuts[0].custom = {"task": "translate"}  # Only matches BLEU
        cuts[0].first_non_padding_cut = cuts[0]

        batch_size = 1
        predictions = torch.randn(batch_size, 100, 1024)
        predictions_lengths = torch.tensor([80])
        predictions_mask = torch.ones(batch_size, 100, dtype=torch.bool)
        targets = torch.randint(0, 1000, (batch_size, 50))
        targets_lengths = torch.tensor([40])
        input_ids = torch.randint(0, 1000, (batch_size, 60))

        mock_multitask_metric.update(
            predictions=predictions,
            predictions_lengths=predictions_lengths,
            predictions_mask=predictions_mask,
            targets=targets,
            targets_lengths=targets_lengths,
            input_ids=input_ids,
            cuts=cuts,
        )

        # WER should be called with empty tensors
        mock_multitask_metric._mock_wer.update.assert_called_once()
        wer_call_args = mock_multitask_metric._mock_wer.update.call_args
        assert wer_call_args.kwargs['predictions'].shape[0] == 0  # Empty

        # BLEU should be called with the one cut
        mock_multitask_metric._mock_bleu.update.assert_called_once()
        bleu_call_args = mock_multitask_metric._mock_bleu.update.call_args
        assert bleu_call_args.kwargs['predictions'].shape[0] == 1


class TestMultiTaskMetricCompute:
    """Test the metric compute functionality"""

    @pytest.mark.unit
    def test_compute_wer_metric(self):
        """Test compute for WER metric"""
        mock_model = Mock()
        mock_model.decoding = Mock()

        cfg = DictConfig(
            {
                "metrics": {
                    "wer": {
                        "constraint": ".task==transcribe",
                        "_target_": "nemo.collections.asr.metrics.WER",
                    }
                }
            }
        )

        with patch('nemo.collections.asr.metrics.multitask.MultiTaskMetric.from_config_dict') as mock_from_config:
            mock_wer = Mock()
            mock_wer.compute.return_value = {
                "val_wer": 0.1,
                "val_wer_num": 10.0,
                "val_wer_denom": 100.0,
            }  # wer, scores, words
            mock_from_config.return_value = mock_wer

            multitask_metric = MultiTaskMetric(mock_model, cfg)

            result = multitask_metric.compute(return_all_metrics=True, prefix="val_")

            expected = {"val_wer": 0.1, "val_wer_num": 10.0, "val_wer_denom": 100.0}
            assert result == expected

    @pytest.mark.unit
    def test_compute_bleu_metric(self):
        """Test compute for BLEU metric"""
        mock_model = Mock()
        mock_model.decoding = Mock()

        cfg = DictConfig(
            {
                "metrics": {
                    "bleu": {
                        "constraint": ".task==translate",
                        "_target_": "nemo.collections.asr.metrics.BLEU",
                    }
                }
            }
        )

        with patch('nemo.collections.asr.metrics.multitask.MultiTaskMetric.from_config_dict') as mock_from_config:
            mock_bleu = Mock()
            mock_bleu.compute.return_value = {"bleu": 25.5, "bleu_num": 1000.0}
            mock_from_config.return_value = mock_bleu

            multitask_metric = MultiTaskMetric(mock_model, cfg)

            result = multitask_metric.compute(prefix="test_")

            assert result == {"bleu": 25.5, "bleu_num": 1000.0}

    @pytest.mark.unit
    def test_reset_metrics(self):
        """Test that reset is called on all metrics"""
        mock_model = Mock()
        mock_model.decoding = Mock()

        cfg = DictConfig(
            {
                "metrics": {
                    "wer": {
                        "constraint": ".task==transcribe",
                        "_target_": "nemo.collections.asr.metrics.WER",
                    },
                    "bleu": {
                        "constraint": ".task==translate",
                        "_target_": "nemo.collections.asr.metrics.BLEU",
                    },
                }
            }
        )

        with patch('nemo.collections.asr.metrics.multitask.MultiTaskMetric.from_config_dict') as mock_from_config:
            mock_wer = Mock()
            mock_bleu = Mock()
            mock_from_config.side_effect = [mock_wer, mock_bleu]

            multitask_metric = MultiTaskMetric(mock_model, cfg)
            multitask_metric.reset()

            mock_wer.reset.assert_called_once()
            mock_bleu.reset.assert_called_once()


class TestMultiTaskMetricEdgeCases:
    """Test edge cases and error conditions"""

    @pytest.mark.unit
    def test_missing_custom_attribute(self):
        """Test handling of cuts without custom attribute"""
        mock_model = Mock()
        mock_model.decoding = Mock()

        cfg = DictConfig(
            {
                "metrics": {
                    "wer": {
                        "constraint": ".task==transcribe",
                        "_target_": "nemo.collections.asr.metrics.WER",
                    }
                }
            }
        )

        with patch('nemo.collections.asr.metrics.multitask.MultiTaskMetric.from_config_dict') as mock_from_config:
            mock_wer = Mock()
            mock_from_config.return_value = mock_wer

            multitask_metric = MultiTaskMetric(mock_model, cfg)

            # Create cut without custom attribute
            cut = Mock()
            cut.custom = {}  # Empty custom dict
            cut.first_non_padding_cut = cut

            cuts_split, idx_split = multitask_metric._split_cuts([cut])

            # Should not match any constraints
            assert idx_split["wer"] == []

    @pytest.mark.unit
    def test_complex_constraint_edge_cases(self):
        """Test complex constraints with edge cases"""
        parser = ConstraintParser()

        # Test constraint with missing properties
        constraint_fn = parser.parse_constraint(".missing_prop==value")
        result = constraint_fn({})
        assert result is False

        # Test constraint with None values
        constraint_fn = parser.parse_constraint(".prop==value")
        result = constraint_fn({"prop": None})
        assert result is False

    @pytest.mark.unit
    def test_operators_coverage(self):
        """Test that all operators are properly defined"""
        parser = ConstraintParser()
        assert "==" in parser.primitives
        assert "!=" in parser.primitives
        assert parser.primitives["=="] == operator.eq
        assert parser.primitives["!="] == operator.ne



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Any, Callable, Dict, Optional, Sequence, Tuple

import nemo_run as run
import torch
import torch.distributed
import torch.utils.checkpoint
import torchvision
import wandb
from autovae import VAEGenerator
from contperceptual_loss import LPIPSWithDiscriminator
from diffusers import AutoencoderKL
from megatron.core import parallel_state
from megatron.core.transformer.enums import ModelType
from megatron.core.transformer.module import MegatronModule
from megatron.core.transformer.transformer_config import TransformerConfig
from megatron.energon import DefaultTaskEncoder, ImageSample
from torch import Tensor, nn

from nemo import lightning as nl
from nemo.collections import llm
from nemo.collections.diffusion.data.diffusion_energon_datamodule import DiffusionDataModule
from nemo.collections.diffusion.train import pretrain
from nemo.collections.llm.gpt.model.base import GPTModel
from nemo.lightning.io.mixin import IOMixin
from nemo.lightning.megatron_parallel import DataT, MegatronLossReduction, ReductionT
from nemo.lightning.pytorch.optim import OptimizerModule


class AvgLossReduction(MegatronLossReduction):
    """Performs average loss reduction across micro-batches."""

    def forward(self, batch: DataT, forward_out: Tensor) -> Tuple[Tensor, ReductionT]:
        """
        Forward pass for loss reduction.

        Args:
            batch: The batch of data.
            forward_out: The output tensor from forward computation.

        Returns:
            A tuple of (loss, reduction dictionary).
        """
        loss = forward_out.mean()
        return loss, {"avg": loss}

    def reduce(self, losses_reduced_per_micro_batch: Sequence[ReductionT]) -> Tensor:
        """
        Reduce losses across multiple micro-batches by averaging them.

        Args:
            losses_reduced_per_micro_batch: A sequence of loss dictionaries.

        Returns:
            The averaged loss tensor.
        """
        losses = torch.stack([loss["avg"] for loss in losses_reduced_per_micro_batch])
        return losses.mean()


class VAE(MegatronModule):
    """Variational Autoencoder (VAE) module."""

    def __init__(self, config, pretrained_model_name_or_path, search_vae=False):
        """
        Initialize the VAE model.

        Args:
            config: Transformer configuration.
            pretrained_model_name_or_path: Path or name of the pretrained model.
            search_vae: Flag to indicate whether to search for a target VAE using AutoVAE.
        """
        super().__init__(config)
        if search_vae:
            # Get VAE automatically from AutoVAE
            self.vae = VAEGenerator(input_resolution=1024, compression_ratio=16)
            # Below line is commented out due to an undefined 'generator' variable in original code snippet.
            # self.vae = generator.search_for_target_vae(parameters_budget=895.178707, cuda_max_mem=0)
        else:
            self.vae = AutoencoderKL.from_config(pretrained_model_name_or_path, weight_dtype=torch.bfloat16)

        sdxl_vae = AutoencoderKL.from_pretrained(
            'stabilityai/stable-diffusion-xl-base-1.0', subfolder="vae", weight_dtype=torch.bfloat16
        )
        sd_dict = sdxl_vae.state_dict()
        vae_dict = self.vae.state_dict()
        pre_dict = {k: v for k, v in sd_dict.items() if (k in vae_dict) and (vae_dict[k].numel() == v.numel())}
        self.vae.load_state_dict(pre_dict, strict=False)
        del sdxl_vae

        self.vae_loss = LPIPSWithDiscriminator(
            disc_start=50001,
            logvar_init=0.0,
            kl_weight=0.000001,
            pixelloss_weight=1.0,
            disc_num_layers=3,
            disc_in_channels=3,
            disc_factor=1.0,
            disc_weight=0.5,
            perceptual_weight=1.0,
            use_actnorm=False,
            disc_conditional=False,
            disc_loss="hinge",
        )

    def forward(self, target, global_step):
        """
        Forward pass through the VAE.

        Args:
            target: Target images.
            global_step: Current global step.

        Returns:
            A tuple (aeloss, log_dict_ae, pred) containing the loss, log dictionary, and predictions.
        """
        posterior = self.vae.encode(target).latent_dist
        z = posterior.sample()
        pred = self.vae.decode(z).sample
        aeloss, log_dict_ae = self.vae_loss(
            inputs=target,
            reconstructions=pred,
            posteriors=posterior,
            optimizer_idx=0,
            global_step=global_step,
            last_layer=self.vae.decoder.conv_out.weight,
        )
        return aeloss, log_dict_ae, pred

    def set_input_tensor(self, input_tensor: Tensor) -> None:
        """
        Set input tensor.

        Args:
            input_tensor: The input tensor to the model.
        """
        pass


class VAEModel(GPTModel):
    """A GPTModel wrapper for the VAE."""

    def __init__(
        self,
        pretrained_model_name_or_path: str,
        optim: Optional[OptimizerModule] = None,
        model_transform: Optional[Callable[[nn.Module], nn.Module]] = None,
    ):
        """
        Initialize the VAEModel.

        Args:
            pretrained_model_name_or_path: Path or name of the pretrained model.
            optim: Optional optimizer module.
            model_transform: Optional function to transform the model.
        """
        self.pretrained_model_name_or_path = pretrained_model_name_or_path
        config = TransformerConfig(num_layers=1, hidden_size=1, num_attention_heads=1)
        self.model_type = ModelType.encoder_or_decoder
        super().__init__(config, optim=optim, model_transform=model_transform)

    def configure_model(self) -> None:
        """Configure the model by initializing the module."""
        if not hasattr(self, "module"):
            self.module = VAE(self.config, self.pretrained_model_name_or_path)

    def data_step(self, dataloader_iter) -> Dict[str, Any]:
        """
        Perform a single data step to fetch a batch from the iterator.

        Args:
            dataloader_iter: The dataloader iterator.

        Returns:
            A dictionary with 'pixel_values' ready for the model.
        """
        batch = next(dataloader_iter)[0]
        return {'pixel_values': batch.image.to(device='cuda', dtype=torch.bfloat16, non_blocking=True)}

    def forward(self, *args, **kwargs):
        """
        Forward pass through the underlying module.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            The result of forward pass of self.module.
        """
        return self.module(*args, **kwargs)

    def training_step(self, batch, batch_idx=None) -> torch.Tensor:
        """
        Perform a single training step.

        Args:
            batch: The input batch.
            batch_idx: Batch index.

        Returns:
            The loss tensor.
        """
        loss, log_dict_ae, pred = self(batch["pixel_values"], self.global_step)

        if torch.distributed.get_rank() == 0:
            self.log_dict(log_dict_ae)

        return loss

    def validation_step(self, batch, batch_idx=None) -> torch.Tensor:
        """
        Perform a single validation step.

        Args:
            batch: The input batch.
            batch_idx: Batch index.

        Returns:
            The loss tensor.
        """
        loss, log_dict_ae, pred = self(batch["pixel_values"], self.global_step)

        image = torch.cat([batch["pixel_values"].cpu(), pred.cpu()], axis=0)
        image = (image + 0.5).clamp(0, 1)

        # wandb is on the last rank for megatron, first rank for nemo
        wandb_rank = 0

        if parallel_state.get_data_parallel_src_rank() == wandb_rank:
            if torch.distributed.get_rank() == wandb_rank:
                gather_list = [None for _ in range(parallel_state.get_data_parallel_world_size())]
            else:
                gather_list = None
            torch.distributed.gather_object(
                image, gather_list, wandb_rank, group=parallel_state.get_data_parallel_group()
            )
            if gather_list is not None:
                self.log_dict(log_dict_ae)
                wandb.log(
                    {
                        "Original (left), Reconstruction (right)": [
                            wandb.Image(torchvision.utils.make_grid(image)) for _, image in enumerate(gather_list)
                        ]
                    },
                )

        return loss

    @property
    def training_loss_reduction(self) -> AvgLossReduction:
        """Returns the loss reduction method for training."""
        if not self._training_loss_reduction:
            self._training_loss_reduction = AvgLossReduction()
        return self._training_loss_reduction

    @property
    def validation_loss_reduction(self) -> AvgLossReduction:
        """Returns the loss reduction method for validation."""
        if not self._validation_loss_reduction:
            self._validation_loss_reduction = AvgLossReduction()
        return self._validation_loss_reduction

    def on_validation_model_zero_grad(self) -> None:
        """
        Hook to handle zero grad on validation model step.
        Used here to skip first validation on resume.
        """
        super().on_validation_model_zero_grad()
        if self.trainer.ckpt_path is not None and getattr(self, '_restarting_skip_val_flag', True):
            self.trainer.sanity_checking = True
            self._restarting_skip_val_flag = False


def crop_image(img, divisor=16):
    """
    Crop the image so that both dimensions are divisible by the given divisor.

    Args:
        img: Image tensor.
        divisor: The divisor to use for cropping.

    Returns:
        The cropped image tensor.
    """
    h, w = img.shape[-2], img.shape[-1]

    delta_h = h % divisor
    delta_w = w % divisor

    delta_h_top = delta_h // 2
    delta_h_bottom = delta_h - delta_h_top

    delta_w_left = delta_w // 2
    delta_w_right = delta_w - delta_w_left

    img_cropped = img[..., delta_h_top : h - delta_h_bottom, delta_w_left : w - delta_w_right]

    return img_cropped


class ImageTaskEncoder(DefaultTaskEncoder, IOMixin):
    """Image task encoder that crops and normalizes the image."""

    def encode_sample(self, sample: ImageSample) -> ImageSample:
        """
        Encode a single image sample by cropping and shifting its values.

        Args:
            sample: An image sample.

        Returns:
            The transformed image sample.
        """
        sample = super().encode_sample(sample)
        sample.image = crop_image(sample.image, 16)
        sample.image -= 0.5
        return sample


@run.cli.factory(target=llm.train)
def train_vae() -> run.Partial:
    """
    Training factory function for VAE.

    Returns:
        A run.Partial recipe for training.
    """
    recipe = pretrain()
    recipe.model = run.Config(
        VAEModel,
        pretrained_model_name_or_path='nemo/collections/diffusion/vae/vae16x/config.json',
    )
    recipe.data = run.Config(
        DiffusionDataModule,
        task_encoder=run.Config(ImageTaskEncoder),
        global_batch_size=24,
        num_workers=10,
    )
    recipe.optim.lr_scheduler = run.Config(nl.lr_scheduler.WarmupHoldPolicyScheduler, warmup_steps=100, hold_steps=1e9)
    recipe.optim.config.lr = 5e-6
    recipe.optim.config.weight_decay = 1e-2
    recipe.log.log_dir = 'nemo_experiments/train_vae'
    recipe.trainer.val_check_interval = 1000
    recipe.trainer.callbacks[0].every_n_train_steps = 1000

    return recipe


if __name__ == "__main__":
    run.cli.main(llm.train, default_factory=train_vae)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os.path
from dataclasses import MISSING, dataclass
from typing import Dict, Optional

from nemo.utils import logging

__all__ = ["get_tokenizer"]


megatron_tokenizer_model_map = {
    "BertWordPieceLowerCase": "megatron-bert-345m-uncased",
    "BertWordPieceCase": "megatron-bert-345m-cased",
    "GPT2BPETokenizer": "megatron-gpt-345m",
}


@dataclass
class TokenizerConfig:
    """
    Tokenizer Configuration Dataclass.
    """

    library: str = MISSING
    tokenizer_model: Optional[str] = None
    vocab_size: Optional[int] = None
    vocab_file: Optional[str] = None
    special_tokens: Optional[Dict[str, str]] = None
    bpe_dropout: Optional[float] = 0.0
    coverage: Optional[float] = 0.999
    training_sample_size: Optional[int] = None
    r2l: Optional[bool] = False
    sentencepiece_legacy: Optional[bool] = False


def get_tokenizer(
    tokenizer_name: str,
    tokenizer_model: Optional[str] = None,
    vocab_file: Optional[str] = None,
    merges_file: Optional[str] = None,
    special_tokens: Optional[Dict[str, str]] = None,
    use_fast: Optional[bool] = False,
    bpe_dropout: Optional[float] = 0.0,
    chat_template: Optional[Dict] = None,
):
    """
    Args:
        tokenizer_name: sentencepiece or pretrained model from the hugging face list,
            for example: bert-base-cased
        tokenizer_model: tokenizer model file of sentencepiece
        special_tokens: dict of special tokens.
            For additional special tokens besides standard special tokens (bos, eos, pad, etc.), such as sentinel
            tokens for T5 (<extra_id_0>, <extra_id_1>, etc.), use key 'additional_special_tokens'
        vocab_file: path to vocab file
        use_fast: (only for HuggingFace AutoTokenizer) set to True to use fast HuggingFace tokenizer
        bpe_dropout: (experimental) BPE dropout tries to corrupt the standard segmentation
            procedure of BPE to help
            model better learn word compositionality and become robust to segmentation errors.
            It has empirically been shown to improve inference time BLEU scores.
    """
    import omegaconf
    from omegaconf import OmegaConf

    if isinstance(
        special_tokens,
        (omegaconf.listconfig.ListConfig, omegaconf.dictconfig.DictConfig),
    ):
        special_tokens = OmegaConf.to_container(special_tokens)

    if special_tokens is None:
        special_tokens_dict = {}
    else:
        special_tokens_dict = special_tokens

    if "megatron" in tokenizer_name:
        try:
            from nemo.collections.common.tokenizers.megatron_utils import (
                get_megatron_merges_file,
                get_megatron_tokenizer,
                get_megatron_vocab_file,
            )
        except (ImportError, ModuleNotFoundError):
            raise ImportError(
                "Megatron-core was not found. Please see the NeMo README for installation instructions: "
                " https://github.com/NVIDIA/NeMo#megatron-gpt."
            )
        if vocab_file is None:
            vocab_file = get_megatron_vocab_file(tokenizer_name)
            merges_file = get_megatron_merges_file(tokenizer_name)
        tokenizer_name = get_megatron_tokenizer(tokenizer_name)

    if tokenizer_name == "sentencepiece":
        from nemo.collections.common.tokenizers.sentencepiece_tokenizer import SentencePieceTokenizer

        logging.info("tokenizer_model: " + str(tokenizer_model))
        return SentencePieceTokenizer(
            model_path=tokenizer_model,
            special_tokens=special_tokens,
            legacy=True,
            chat_template=chat_template,
        )
    elif tokenizer_name == "tiktoken":
        from nemo.collections.common.tokenizers.tiktoken_tokenizer import TiktokenTokenizer

        return TiktokenTokenizer(
            vocab_file=vocab_file,
            special_tokens=special_tokens["additional_special_tokens"],
        )
    elif tokenizer_name == "word":
        from nemo.collections.common.tokenizers.word_tokenizer import WordTokenizer

        return WordTokenizer(vocab_file=vocab_file, **special_tokens_dict)
    elif tokenizer_name == "char":
        from nemo.collections.common.tokenizers.char_tokenizer import CharTokenizer

        return CharTokenizer(vocab_file=vocab_file, **special_tokens_dict)
    elif tokenizer_name == "regex":
        from nemo.collections.common.tokenizers.regex_tokenizer import RegExTokenizer

        return RegExTokenizer().load_tokenizer(regex_file=tokenizer_model, vocab_file=vocab_file)

    logging.info(
        f"Getting HuggingFace AutoTokenizer with pretrained_model_name: {tokenizer_name}, vocab_file: {vocab_file}, "
        f" merges_files: {merges_file}, special_tokens_dict: {special_tokens_dict}, and use_fast: {use_fast}"
    )
    from nemo.collections.common.tokenizers.huggingface.auto_tokenizer import AutoTokenizer

    tokenizer = AutoTokenizer(
        pretrained_model_name=tokenizer_name,
        vocab_file=vocab_file,
        merges_file=merges_file,
        **special_tokens_dict,
        use_fast=use_fast,
        chat_template=chat_template,
    )
    return tokenizer


def get_nmt_tokenizer(
    library: str = "sentencepiece",
    model_name: Optional[str] = None,
    tokenizer_model: Optional[str] = None,
    vocab_file: Optional[str] = None,
    merges_file: Optional[str] = None,
    special_tokens: Optional[Dict[str, str]] = None,
    use_fast: Optional[bool] = False,
    bpe_dropout: Optional[float] = 0.0,
    r2l: Optional[bool] = False,
    legacy: Optional[bool] = False,
    delimiter: Optional[str] = None,
    trust_remote_code: Optional[bool] = False,
    chat_template: Optional[Dict] = None,
    vocab_size: Optional[int] = None,
):
    """
    Args:
        model_name: if using a pretrained model from NeMo, HuggingFace, or Megatron
        tokenizer_model: tokenizer model file of sentencepiece
        special_tokens: dict of special tokens
        vocab_file: path to vocab file
        use_fast: (only for HuggingFace AutoTokenizer) set to True to use fast HuggingFace tokenizer
        bpe_dropout: (experimental) BPE dropout tries to corrupt the standard segmentation procedure
            of BPE to help model better learn word compositionality and become robust to segmentation errors.
            It has empirically been shown to improve inference time BLEU scores.
        r2l: Whether to return subword IDs from right to left
    """
    import omegaconf
    from omegaconf import OmegaConf

    if isinstance(
        special_tokens,
        (omegaconf.listconfig.ListConfig, omegaconf.dictconfig.DictConfig),
    ):
        special_tokens = OmegaConf.to_container(special_tokens)
    if special_tokens is None:
        special_tokens_dict = {}
    else:
        special_tokens_dict = special_tokens

    if (library != "byte-level") and (
        model_name is None and (tokenizer_model is None or not os.path.isfile(tokenizer_model))
    ):
        raise ValueError("No Tokenizer path provided or file does not exist!")

    if library == "huggingface":
        from nemo.collections.common.tokenizers.huggingface.auto_tokenizer import AutoTokenizer

        logging.info(f'Getting HuggingFace AutoTokenizer with pretrained_model_name: {model_name}')
        tokenizer = AutoTokenizer(
            pretrained_model_name=model_name,
            vocab_file=vocab_file,
            merges_file=merges_file,
            **special_tokens_dict,
            use_fast=use_fast,
            trust_remote_code=trust_remote_code,
            chat_template=chat_template,
        )
        if chat_template:
            tokenizer.tokenizer.chat_template = chat_template
        return tokenizer
    elif library == 'sentencepiece':
        from nemo.collections.common.tokenizers.sentencepiece_tokenizer import SentencePieceTokenizer

        logging.info(f"Getting SentencePiece with model: {tokenizer_model}")

        return SentencePieceTokenizer(
            model_path=tokenizer_model,
            special_tokens=special_tokens,
            legacy=legacy,
            chat_template=chat_template,
        )
    elif library == "byte-level":
        from nemo.collections.common.tokenizers.bytelevel_tokenizers import ByteLevelTokenizer

        logging.info("Using byte-level tokenization")
        return ByteLevelTokenizer(special_tokens_dict)
    elif library == "regex":
        from nemo.collections.common.tokenizers.regex_tokenizer import RegExTokenizer

        logging.info("Using regex tokenization")
        return RegExTokenizer().load_tokenizer(regex_file=tokenizer_model, vocab_file=vocab_file)
    elif library == "megatron":
        if model_name == "GPTSentencePieceTokenizer":
            from nemo.collections.common.tokenizers.sentencepiece_tokenizer import SentencePieceTokenizer

            logging.info("tokenizer_model: ")
            logging.info(tokenizer_model)
            return SentencePieceTokenizer(model_path=tokenizer_model, legacy=legacy)

        if model_name in megatron_tokenizer_model_map:
            model_name = megatron_tokenizer_model_map[model_name]
        logging.info(
            f"Getting Megatron tokenizer for pretrained model name: {model_name}, custom vocab file: {vocab_file}, "
            f"and merges file: {merges_file}"
        )
        return get_tokenizer(
            tokenizer_name=model_name,
            vocab_file=vocab_file,
            merges_file=merges_file,
            special_tokens=special_tokens_dict,
            chat_template=chat_template,
        )
    elif library == "tabular":
        from nemo.collections.common.tokenizers.tabular_tokenizer import TabularTokenizer

        return TabularTokenizer(vocab_file, delimiter=delimiter)
    elif library == "tiktoken":
        from nemo.collections.common.tokenizers.tiktoken_tokenizer import TiktokenTokenizer

        return TiktokenTokenizer(vocab_file=vocab_file)
    elif library == "null":
        assert vocab_size is not None
        from nemo.collections.common.tokenizers.null_tokenizer import NullTokenizer

        return NullTokenizer(vocab_size)
    else:
        raise NotImplementedError(
            'Currently we only support "huggingface", "sentencepiece", "megatron", "byte-level", "regex", "tabular",'
            '"tiktoken", and "null" tokenizer libraries.'
        )



# Copyright (c) 2022, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
from datetime import date, datetime
from decimal import Decimal
from unittest import mock

import pytest

from nemo.utils.env_var_parsing import (
    CoercionError,
    RequiredSettingMissingError,
    get_envbool,
    get_envdate,
    get_envdatetime,
    get_envdecimal,
    get_envdict,
    get_envfloat,
    get_envint,
    get_envlist,
)


class TestEnvironmentVariableParsing:

    def test_get_envint_returns_int_value(self):
        """Test that get_envint returns the integer value from environment variable."""
        with mock.patch.dict(os.environ, {'TEST_INT': '42'}):
            assert get_envint('TEST_INT') == 42

    def test_get_envint_with_default(self):
        """Test that get_envint returns the default value when env var is missing."""
        assert get_envint('NONEXISTENT_INT', 123) == 123

    def test_get_envint_required_missing(self):
        """Test that get_envint raises an exception when required env var is missing."""
        with pytest.raises(RequiredSettingMissingError):
            get_envint('NONEXISTENT_INT')

    def test_get_envint_coercion_error(self):
        """Test that get_envint raises a CoercionError for non-integer values."""
        with mock.patch.dict(os.environ, {'TEST_INT': 'not-an-int'}):
            with pytest.raises(CoercionError):
                get_envint('TEST_INT')

    def test_get_envint_negative_value(self):
        """Test that get_envint correctly handles negative integers."""
        with mock.patch.dict(os.environ, {'TEST_INT': '-42'}):
            assert get_envint('TEST_INT') == -42

    def test_get_envfloat_returns_float_value(self):
        """Test that get_envfloat returns the float value from environment variable."""
        with mock.patch.dict(os.environ, {'TEST_FLOAT': '3.14'}):
            assert get_envfloat('TEST_FLOAT') == 3.14

    def test_get_envfloat_with_integer_string(self):
        """Test that get_envfloat correctly converts integer strings to floats."""
        with mock.patch.dict(os.environ, {'TEST_FLOAT': '42'}):
            value = get_envfloat('TEST_FLOAT')
            assert value == 42.0
            assert isinstance(value, float)

    def test_get_envfloat_with_default(self):
        """Test that get_envfloat returns the default value when env var is missing."""
        assert get_envfloat('NONEXISTENT_FLOAT', 3.14) == 3.14

    def test_get_envfloat_required_missing(self):
        """Test that get_envfloat raises an exception when required env var is missing."""
        with pytest.raises(RequiredSettingMissingError):
            get_envfloat('NONEXISTENT_FLOAT')

    def test_get_envfloat_coercion_error(self):
        """Test that get_envfloat raises a CoercionError for non-float values."""
        with mock.patch.dict(os.environ, {'TEST_FLOAT': 'not-a-float'}):
            with pytest.raises(CoercionError):
                get_envfloat('TEST_FLOAT')

    def test_get_envfloat_scientific_notation(self):
        """Test that get_envfloat correctly handles scientific notation."""
        with mock.patch.dict(os.environ, {'TEST_FLOAT': '1.23e-4'}):
            assert get_envfloat('TEST_FLOAT') == 1.23e-4

    def test_get_envfloat_negative_value(self):
        """Test that get_envfloat correctly handles negative values."""
        with mock.patch.dict(os.environ, {'TEST_FLOAT': '-3.14'}):
            assert get_envfloat('TEST_FLOAT') == -3.14

    # Tests for get_envbool
    def test_get_envbool_true_values(self):
        """Test that get_envbool returns True for various truthy values."""
        true_values = ['true', 'True', 'TRUE', '1', 'yes', 'Yes', 'YES', 'y', 'Y', 't', 'T']
        for val in true_values:
            with mock.patch.dict(os.environ, {'TEST_BOOL': val}):
                assert get_envbool('TEST_BOOL') is True

    def test_get_envbool_false_values(self):
        """Test that get_envbool returns False for various falsy values."""
        false_values = ['false', 'False', 'FALSE', '0', 'no', 'No', 'NO', 'n', 'N', 'f', 'F', 'none', 'None', 'NONE']
        for val in false_values:
            with mock.patch.dict(os.environ, {'TEST_BOOL': val}):
                assert get_envbool('TEST_BOOL') is False

    def test_get_envbool_with_default(self):
        """Test that get_envbool returns the default value when env var is missing."""
        assert get_envbool('NONEXISTENT_BOOL', True) is True
        assert get_envbool('NONEXISTENT_BOOL', False) is False

    def test_get_envbool_required_missing(self):
        """Test that get_envbool raises an exception when required env var is missing."""
        with pytest.raises(RequiredSettingMissingError):
            get_envbool('NONEXISTENT_BOOL')

    def test_get_envbool_non_boolean_value(self):
        """Test that get_envbool interprets non-standard values as True."""
        with mock.patch.dict(os.environ, {'TEST_BOOL': 'something-else'}):
            assert get_envbool('TEST_BOOL') is True

    # Tests for get_envdecimal
    def test_get_envdecimal_returns_decimal_value(self):
        """Test that get_envdecimal returns the Decimal value from environment variable."""
        with mock.patch.dict(os.environ, {'TEST_DECIMAL': '3.14'}):
            value = get_envdecimal('TEST_DECIMAL')
            assert value == Decimal('3.14')
            assert isinstance(value, Decimal)

    def test_get_envdecimal_with_integer_string(self):
        """Test that get_envdecimal correctly converts integer strings to Decimals."""
        with mock.patch.dict(os.environ, {'TEST_DECIMAL': '42'}):
            value = get_envdecimal('TEST_DECIMAL')
            assert value == Decimal('42')
            assert isinstance(value, Decimal)

    def test_get_envdecimal_with_default(self):
        """Test that get_envdecimal returns the default value when env var is missing."""
        default_value = Decimal('3.14')
        assert get_envdecimal('NONEXISTENT_DECIMAL', default_value) == default_value

    def test_get_envdecimal_required_missing(self):
        """Test that get_envdecimal raises an exception when required env var is missing."""
        with pytest.raises(RequiredSettingMissingError):
            get_envdecimal('NONEXISTENT_DECIMAL')

    def test_get_envdecimal_coercion_error(self):
        """Test that get_envdecimal raises a CoercionError for non-decimal values."""
        with mock.patch.dict(os.environ, {'TEST_DECIMAL': 'not-a-decimal'}):
            with pytest.raises(CoercionError):
                get_envdecimal('TEST_DECIMAL')

    def test_get_envdecimal_negative_value(self):
        """Test that get_envdecimal correctly handles negative values."""
        with mock.patch.dict(os.environ, {'TEST_DECIMAL': '-3.14'}):
            assert get_envdecimal('TEST_DECIMAL') == Decimal('-3.14')

    def test_get_envdecimal_high_precision(self):
        """Test that get_envdecimal preserves high precision values."""
        with mock.patch.dict(os.environ, {'TEST_DECIMAL': '3.1415926535897932384626433832795028841971'}):
            value = get_envdecimal('TEST_DECIMAL')
            assert value == Decimal('3.1415926535897932384626433832795028841971')

    # Tests for get_envdate
    def test_get_envdate_returns_date_value(self):
        """Test that get_envdate returns the date value from environment variable."""
        with mock.patch.dict(os.environ, {'TEST_DATE': '2023-05-15'}):
            value = get_envdate('TEST_DATE')
            assert value == date(2023, 5, 15)
            assert isinstance(value, date)

    def test_get_envdate_with_different_formats(self):
        """Test that get_envdate handles different date formats."""
        date_formats = {
            '2023-05-15': date(2023, 5, 15),
            '15-05-2023': date(2023, 5, 15),
            '05/15/2023': date(2023, 5, 15),
            '15 May 2023': date(2023, 5, 15),
            'May 15, 2023': date(2023, 5, 15),
        }

        for date_str, expected_date in date_formats.items():
            with mock.patch.dict(os.environ, {'TEST_DATE': date_str}):
                assert get_envdate('TEST_DATE') == expected_date

    def test_get_envdate_with_default(self):
        """Test that get_envdate returns the default value when env var is missing."""
        default_value = date(2023, 5, 15)
        assert get_envdate('NONEXISTENT_DATE', default_value) == default_value

    def test_get_envdate_required_missing(self):
        """Test that get_envdate raises an exception when required env var is missing."""
        with pytest.raises(RequiredSettingMissingError):
            get_envdate('NONEXISTENT_DATE')

    def test_get_envdate_coercion_error(self):
        """Test that get_envdate raises a CoercionError for invalid date values."""
        with mock.patch.dict(os.environ, {'TEST_DATE': 'not-a-date'}):
            with pytest.raises(CoercionError):
                get_envdate('TEST_DATE')

    # Tests for get_envdatetime
    def test_get_envdatetime_returns_datetime_value(self):
        """Test that get_envdatetime returns the datetime value from environment variable."""
        with mock.patch.dict(os.environ, {'TEST_DATETIME': '2023-05-15T14:30:45'}):
            value = get_envdatetime('TEST_DATETIME')
            assert value == datetime(2023, 5, 15, 14, 30, 45)
            assert isinstance(value, datetime)

    def test_get_envdatetime_with_different_formats(self):
        """Test that get_envdatetime handles different datetime formats."""
        datetime_formats = {
            '2023-05-15T14:30:45': datetime(2023, 5, 15, 14, 30, 45),
            '2023-05-15 14:30:45': datetime(2023, 5, 15, 14, 30, 45),
            '15-05-2023 14:30:45': datetime(2023, 5, 15, 14, 30, 45),
            '05/15/2023 14:30:45': datetime(2023, 5, 15, 14, 30, 45),
            '15 May 2023 14:30:45': datetime(2023, 5, 15, 14, 30, 45),
        }

        for datetime_str, expected_datetime in datetime_formats.items():
            with mock.patch.dict(os.environ, {'TEST_DATETIME': datetime_str}):
                assert get_envdatetime('TEST_DATETIME') == expected_datetime

    def test_get_envdatetime_with_default(self):
        """Test that get_envdatetime returns the default value when env var is missing."""
        default_value = datetime(2023, 5, 15, 14, 30, 45)
        assert get_envdatetime('NONEXISTENT_DATETIME', default_value) == default_value

    def test_get_envdatetime_required_missing(self):
        """Test that get_envdatetime raises an exception when required env var is missing."""
        with pytest.raises(RequiredSettingMissingError):
            get_envdatetime('NONEXISTENT_DATETIME')

    def test_get_envdatetime_coercion_error(self):
        """Test that get_envdatetime raises a CoercionError for invalid datetime values."""
        with mock.patch.dict(os.environ, {'TEST_DATETIME': 'not-a-datetime'}):
            with pytest.raises(CoercionError):
                get_envdatetime('TEST_DATETIME')

    def test_get_envdatetime_with_timezone(self):
        """Test that get_envdatetime handles timezone information."""
        with mock.patch.dict(os.environ, {'TEST_DATETIME': '2023-05-15T14:30:45+0200'}):
            dt = get_envdatetime('TEST_DATETIME')
            assert dt.year == 2023
            assert dt.month == 5
            assert dt.day == 15
            assert dt.hour == 14
            assert dt.minute == 30
            assert dt.second == 45

    # Tests for get_envlist
    def test_get_envlist_returns_list_value(self):
        """Test that get_envlist returns the list value from environment variable."""
        with mock.patch.dict(os.environ, {'TEST_LIST': 'item1 item2 item3'}):
            value = get_envlist('TEST_LIST')
            assert value == ['item1', 'item2', 'item3']
            assert isinstance(value, list)

    def test_get_envlist_with_custom_separator(self):
        """Test that get_envlist handles custom separators."""
        with mock.patch.dict(os.environ, {'TEST_LIST': 'item1,item2,item3'}):
            value = get_envlist('TEST_LIST', separator=',')
            assert value == ['item1', 'item2', 'item3']

    def test_get_envlist_with_default(self):
        """Test that get_envlist returns the default value when env var is missing."""
        default_value = ['default1', 'default2']
        assert get_envlist('NONEXISTENT_LIST', default_value) == default_value

    def test_get_envlist_required_missing(self):
        """Test that get_envlist raises an exception when required env var is missing."""
        with pytest.raises(RequiredSettingMissingError):
            get_envlist('NONEXISTENT_LIST')

    def test_get_envlist_empty_string(self):
        """Test that get_envlist handles empty strings."""
        with mock.patch.dict(os.environ, {'TEST_LIST': ''}):
            value = get_envlist('TEST_LIST')
            assert value == ['']

    def test_get_envlist_multiple_words(self):
        """Test that get_envlist correctly splits words with spaces."""
        with mock.patch.dict(os.environ, {'TEST_LIST': 'word1 "phrase with spaces" word3'}):
            value = get_envlist('TEST_LIST')
            assert value == ['word1', '"phrase', 'with', 'spaces"', 'word3']

    # Tests for get_envdict
    def test_get_envdict_returns_dict_value(self):
        """Test that get_envdict returns the dict value from environment variable."""
        test_dict = {'key1': 'value1', 'key2': 42, 'key3': True}
        with mock.patch.dict(os.environ, {'TEST_DICT': json.dumps(test_dict)}):
            value = get_envdict('TEST_DICT')
            assert value == test_dict
            assert isinstance(value, dict)

    def test_get_envdict_with_default(self):
        """Test that get_envdict returns the default value when env var is missing."""
        default_value = {'default_key': 'default_value'}
        assert get_envdict('NONEXISTENT_DICT', default_value) == default_value

    def test_get_envdict_required_missing(self):
        """Test that get_envdict raises an exception when required env var is missing."""
        with pytest.raises(RequiredSettingMissingError):
            get_envdict('NONEXISTENT_DICT')

    def test_get_envdict_coercion_error(self):
        """Test that get_envdict raises a CoercionError for invalid JSON."""
        with mock.patch.dict(os.environ, {'TEST_DICT': 'not-valid-json'}):
            with pytest.raises(CoercionError):
                get_envdict('TEST_DICT')

    def test_get_envdict_complex_dict(self):
        """Test that get_envdict handles complex nested dictionaries."""
        complex_dict = {
            'string': 'value',
            'number': 42,
            'boolean': True,
            'list': [1, 2, 3],
            'nested': {'key1': 'value1', 'key2': [4, 5, 6]},
        }
        with mock.patch.dict(os.environ, {'TEST_DICT': json.dumps(complex_dict)}):
            value = get_envdict('TEST_DICT')
            assert value == complex_dict
